
/***************************************************************************
 *   Copyright (C) 2009 by Jaime Pulido y Salvador DomÃ­nguez   *
 *   saldom@cartif.es   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/**
  @file eusebio.cpp
  @brief ROS node to interact with the robot platform
  @author Salvador Dominguez
  @date 13/8/2013
  */

#include "robot.h"

//The robot object
Robot rb;
//Parameters file names
string robot_status_filename,charging_pose_filename;

/**
  @brief Callback for the ROS joy msg
  @param msg ROS joy msg
  */
//void JoyHnd(joy::Joy msg){
void JoyHnd(sensor_msgs::Joy msg){
  //Check the number of buttons and axes received
  if ((msg.buttons.size()<MIN_NUM_JOY_BUTTONS)|| (msg.axes.size()<MIN_NUM_JOY_AXES)) {
      printf("NUMBER OF JOY BUTTONS OR AXES TOO LOW\n");
      return;
  }


  //Reset the dig control flags vector
  rb.robot_st.dig_ctr_state.clear();

  //Dig control flags vector updating according to the received joy buttons
  for (unsigned int i=0;i<msg.buttons.size();i++){
    rb.robot_st.dig_ctr_state.push_back(msg.buttons[i]);
    //printf("%d\t",msg.buttons[i]);
  }

  //Dead man button condition
  rb.dead_man_button=msg.buttons[JOY_PC_BUTT_L2];

  //Joy X Y  axes values
  if (msg.axes[JOY_PC_BUTT_LEFT_RIGHT]) rb.ref_joy_x=(double)msg.axes[JOY_PC_BUTT_LEFT_RIGHT];
  else rb.ref_joy_x=(double)msg.axes[JOY_PC_AXIS_X1];
  if (msg.axes[JOY_PC_BUTT_UP_DOWN]) rb.ref_joy_y=(double)msg.axes[JOY_PC_BUTT_UP_DOWN];
  else rb.ref_joy_y=(double)msg.axes[JOY_PC_AXIS_Y1];
  //printf("JOY_X:%f JOY_Y:%f\n",rb.ref_joy_x,rb.ref_joy_y);

  //New joy data available
  rb.new_joy_data=true;



  //Update buttons status
  rb.R2_button=msg.buttons[JOY_PC_BUTT_R2];
  rb.cross_button=msg.buttons[JOY_PC_BUTT_CROSS];
  rb.square_button=msg.buttons[JOY_PC_BUTT_SQUARE];
  rb.circle_button=msg.buttons[JOY_PC_BUTT_CIRCLE];
  rb.triangle_button=msg.buttons[JOY_PC_BUTT_TRIANGLE];
  rb.home_button=msg.buttons[JOY_PC_BUTT_HOME];
  rb.start_button=msg.buttons[JOY_PC_BUTT_START];
}


void Move(){

int16_t ref1_joy=0,ref2_joy=0;

if (rb.dead_man_button==true){

    if (rb.bmp_front && rb.ref_joy_x>0.0) {
        printf("BUMPER FRONT\n");
        rb.ref_joy_x=0.0;
        rb.ref_joy_y=0.0;
    }else if (rb.bmp_back && rb.ref_joy_x<0.0) {
        printf("BUMPER REAR\n");
        rb.ref_joy_x=0.0;
        rb.ref_joy_y=0.0;
    }

    //Send back references to the GPRMC node
    rb.TwistToTracks(rb.ref_joy_x,rb.ref_joy_y,&ref1_joy,&ref2_joy);
    //printf("LOCAL_JOY_CONTROL: RefX:%f RefY:%f Axis1_ref:%d Axis2_ref:%d\n",rb.ref_joy_x,rb.ref_joy_y,ref1_joy,ref2_joy);

    //printf("EUSEBIO:Publishing references\n");
    rb.CtrlRobot(ref1_joy,ref2_joy,rb.robot_st.dig_ctr_state);


/*
         if (rb.bmp_front && ref_lin<0.0) {
             printf("BUMPER FRONT\n");
             rb.CtrlRobot(ref1_joy,ref2_joy,rb.robot_st.dig_ctr_state);
         }else if (rb.bmp_back && ref_lin>0.0) {
             printf("BUMPER REAR\n");
             rb.MoveRobot(0,0);
         }else if (ref_lin<0){//Forward movement
           if (rb.dist_obstacle<SAFE_DIST_RANGE_FRONT){
             printf("OBSTACLE TOO CLOSED IN FRONT. Refang:%d\n",ref_ang);
             rb.TwistToTracks(0.0,0.0,&ref1_joy,&ref2_joy);
             rb.MoveRobot(ref1_joy,ref2_joy);
           }else if (rb.obstacle==-1) {
             printf("OBSTACLE DETECTED, LEFT SIDE\n");
             rb.TwistToTracks(0.0,-(ANG_VEL_SEARCH*VEL_JOY_ANG_COEF),&ref1_joy,&ref2_joy);
             rb.MoveRobot(ref1_joy,ref2_joy);
           }else if (rb.obstacle==1) {
             printf("OBSTACLE DETECTED, RIGHT SIDE\n");
             rb.TwistToTracks(0.0,(ANG_VEL_SEARCH*VEL_JOY_ANG_COEF),&ref1_joy,&ref2_joy);
             rb.MoveRobot(ref1_joy,ref2_joy);
           }else if (rb.dist_obstacle<SAFE_DIST_RANGE_FRONT2){
             ref_lin=(int)((double)ref_lin*(rb.dist_obstacle-SAFE_DIST_RANGE_FRONT)/(SAFE_DIST_RANGE_FRONT2-SAFE_DIST_RANGE_FRONT1));
             printf("OBSTACLE NOT TOO CLOSED: Speed:%d\n",ref_lin);
             rb.TwistToTracks(ref_lin,ref_ang,&ref1_joy,&ref2_joy);
             rb.MoveRobot(ref1_joy,ref2_joy);
           }else{
             printf("NO OBSTACLE\n");
             rb.TwistToTracks(ref_lin,ref_ang,&ref1_joy,&ref2_joy);
             rb.MoveRobot(ref1_joy,ref2_joy);
           }
         }else{//Backward movement
           printf("BACKWARD MOVEMENT: Speed:%d\n",ref_lin);
           rb.TwistToTracks(ref_lin,ref_ang,&ref1_joy,&ref2_joy);
           rb.MoveRobot(ref1_joy,ref2_joy);
         }
*/

    rb.dead_man_button=false;
}//else printf("NO JOY CONTROL\n");


}
/**
  @brief Callback for the ros GPMRC status msg
  @param msg ROS gpmrc_state msg
  */
void GPMRCStateHnd(gpmrc_node::gpmrc_state msg){
    gpmrc_node::AxisData axisLeft,axisRight;
    static double total_dist=0.0;
    static gpmrc_node::gpmrc_state last_msg;
    static bool init=true;
    static double t_ant=0.0, t_set_ch_pose=0.0;
    double t_now;
    int16_t ref1_joy=0,ref2_joy=0;
    std_msgs::String msg_command;
    double curvature;
    double deltaTheta;
    //double qx,qy,qz,qw;
    double rot[9];
    double odom_transf[16];
    double pitch,yaw,roll;
    static tf::TransformBroadcaster br;
    geometry_msgs::TransformStamped out_transform;
    int incr_pls_left,incr_pls_right;
    double incr_dist_left,incr_dist_right;

    //printf("EUSEBIO:Receiving GPMRC state\n");

    //Only process if we have a new motor reading
    if (msg.mot_read==false){
       return;
    }

    t_now=(double)(msg.header.stamp.sec+msg.header.stamp.nsec*1E-9);
    if (init==true) {
        last_msg=msg;
        init=false;
        rb.incr_t=0.0;
    }else{
        rb.incr_t=t_now-t_ant;
    }

    t_ant=t_now;

    rb.timestamp=msg.header.stamp;


    incr_pls_left=msg.axes[0].odo-rb.incr_pls_left_last;
    incr_pls_right=(msg.axes[1].odo-rb.incr_pls_right_last);

   // printf("IncrDist: Left:%d Right:%d Odo: Left:%d Right:%d\n",incr_pls_left,incr_pls_right,msg.axes[0].odo,-msg.axes[1].odo);
    if(abs(incr_pls_left)>200)incr_pls_left=0;
    if(abs(incr_pls_right)>200)incr_pls_right=0;
    //printf("IncrDist: Left:%d Right:%d Odo: Left:%d Right:%d\n",incr_pls_left,incr_pls_right,msg.axes[0].odo,-msg.axes[1].odo);
    rb.incr_pls_left_last=msg.axes[0].odo;
    rb.incr_pls_right_last=msg.axes[1].odo;

    incr_pls_right=-incr_pls_right;

 /*   incr_pls_left=msg.axes[0].vel;
    incr_pls_right=-msg.axes[1].vel;
*/


    //Increment of movement
    rb.incr_dist_robot=((double)(incr_pls_left+incr_pls_right)/2.0)/rb.pls_dist;

    incr_dist_left=(double)(incr_pls_left)/rb.pls_dist;
    incr_dist_right=(double)(incr_pls_right)/rb.pls_dist;
    //rb.incr_dist_robot=((double)(incr_pls_left+incr_pls_right)/2.0);
    printf("Incr_dist_robot:%f incr_dist_left:%f incr_dist_right:%f\n",rb.incr_dist_robot,incr_dist_left,incr_dist_right);
    total_dist+=(double)rb.incr_dist_robot;
    //printf("DIST TRAVELLED: Odo(m):%f\n",total_dist);

    //----------------- ODOMETRY --------------
    if (incr_pls_right!=incr_pls_left){
      double xc,yc;
      double rot_radius;
/*
      curvature=(2.0/rb.dist_between_wheels)*((double)(incr_pls_right-incr_pls_left))/(incr_pls_right+incr_pls_left);
      xc=rb.xs_odom-sin(rb.theta_odom)/curvature;
      yc=rb.ys_odom+cos(rb.theta_odom)/curvature;            

      deltaTheta=rb.incr_dist_robot*curvature;
      rb.xs_odom=xc+sin(rb.theta_odom+deltaTheta)/curvature;
      rb.ys_odom=yc-cos(rb.theta_odom+deltaTheta)/curvature;
      rb.theta_odom+=deltaTheta;
      if (rb.theta_odom > M_PI)rb.theta_odom-=2.0*M_PI;
      else if (rb.theta_odom <= -M_PI) rb.theta_odom+=2.0*M_PI;

     // printf("Curvature:%f Incr_dist:%f DeltaTheta:%f\n",curvature,rb.incr_dist_robot,deltaTheta*180.0/M_PI);
    }else if (rb.incr_dist_robot){
      //printf("Incr_dist:%f\n",rb.incr_dist_robot);
      deltaTheta=0.0;
      rb.xs_odom+=rb.incr_dist_robot*cos(rb.theta_odom);
      rb.ys_odom+=rb.incr_dist_robot*sin(rb.theta_odom);
      rb.theta_odom+=deltaTheta;
      if (rb.theta_odom > M_PI) {
        while (rb.theta_odom > M_PI) rb.theta_odom-=2.0*M_PI;
      }else if (rb.theta_odom <= -M_PI) {
        while (rb.theta_odom <= -M_PI) rb.theta_odom+=2.0*M_PI;
      }
    }
*/
        rb.incr_ang=(double)(incr_dist_right-incr_dist_left)/rb.dist_between_wheels;
        rot_radius=rb.incr_dist_robot/rb.incr_ang;
        rb.incr_x=rot_radius*(cos(rb.incr_ang)-1.0);
        rb.incr_y=rot_radius*sin(rb.incr_ang);

    }else if (rb.incr_dist_robot){
        rb.incr_x=0.0;
        rb.incr_y=rb.incr_dist_robot;
      }else {
        rb.incr_x=rb.incr_y=rb.incr_ang=0.0;
      }
    rb.xs_odom+=rb.incr_x;
    rb.ys_odom+=rb.incr_y;
    rb.theta_odom+=rb.incr_ang;
    if (rb.theta_odom > M_PI)rb.theta_odom-=2.0*M_PI;
    else if (rb.theta_odom <= -M_PI) rb.theta_odom+=2.0*M_PI;

    printf("rb.incr_x:%f rb.incr_y:%f rb.incr_ang:%f\n",rb.incr_x,rb.incr_x,rb.incr_ang);
    //SONARS
    {
        if (msg.son_values.size()){
          int dist_safe;
          bool obst_left,obst_right;
          int dist_object_lat;
          int dist_object_front;

          //printf("SONARS READINGS RECEIVED: 0:%d 1:%d 2:%d 3:%d\n",msg.son_values[0],msg.son_values[1],msg.son_values[2],msg.son_values[3]);

          dist_object_lat=(int)(SAFE_DIST_RANGE_LAT*100.0);
          dist_object_front=(int)(SAFE_DIST_RANGE_FRONT*100.0);
          dist_safe=(int)(rb.incr_dist_robot*100.0);
          //printf("Dist lat:%d, Dist front:%d\n",dist_object_lat,dist_object_front);
          //printf("Safe dist:%f\n",dist_safe);

/*
          if (msg.son_values[1]<msg.son_values[2]) rb.dist_obstacle=(double)msg.son_values[1]/100.0;
          else rb.dist_obstacle=(double)msg.son_values[2]/100.0;
          printf("Dist1:%d Dist2:%d DIST OBSTACLE:%f\n",msg.son_values[1],msg.son_values[2],rb.dist_obstacle);

          if (msg.son_values[0]<(dist_object_lat+dist_safe)) obst_left=true;
          else obst_left=false;
          if (msg.son_values[3]<(dist_object_lat+dist_safe)) obst_right=true;
          else obst_right=false;

          if ((obst_left==true)&&(obst_right==false)){
              //printf("OBSTACLE DETECTED, LEFT SIDE\n");
              rb.son_obstacle=-1;
          }else if ((obst_right==true)&&(obst_left==false)){
              //printf("OBSTACLE DETECTED, RIGHT SIDE\n");
              rb.son_obstacle=1;
          }else {
              //printf("NO OBSTACLE\n");
              rb.son_obstacle=0;
          }
          */
      }
    }
    //printf("WITHOUT CORR: X:%f Y:%f Ori:%f\n",rb.xs_odom,rb.ys_odom,rb.theta_odom*180.0/M_PI);

    //Odometry transform update
    rb.mt.ResetTransform(&odom_transf[0]);

    rb.mt.FromEulerToRotationMatrixYXZ(0.0,0.0,rb.theta_odom,&rot[0]);

    rb.mt.setRotMatrixInTransf(&rot[0],&odom_transf[0]);
    odom_transf[3]=rb.xs_odom;
    odom_transf[7]=rb.ys_odom;
    odom_transf[11]=0.0;

    //printf("CORR: X:%f Y:%f\n",rb.correction_transf[3],rb.correction_transf[7]);
    for (int i=0;i<16;i++) rb.global_transf[i]=rb.correction_transf[i];

    rb.mt.updateTransform(&rb.global_transf[0],&odom_transf[0]);
    rb.mt.getRotMatrixFromTransf(&rb.global_transf[0],&rot[0]);
    rb.mt.FromRotationToQuaternion(&rot[0],&rb.qx,&rb.qy,&rb.qz,&rb.qw );
    rb.mt.FromRotationMatrixToEulerYXZ(&rot[0],&pitch,&yaw,&roll);
    //printf("Pitch:%f Yaw:%f Roll:%f\n",pitch*180.0/M_PI,yaw*180.0/M_PI,roll*180.0/M_PI);
    rb.robot_st.theta=(float)roll;
    rb.robot_st.xs=rb.global_transf[3];
    rb.robot_st.ys=rb.global_transf[7];

    //printf("CORRECTED POSITION: X:%f Y:%f Ori:%f\n",rb.robot_st.xs,rb.robot_st.ys,rb.robot_st.theta*180.0/M_PI);

    //printf("Robot: INCR_ANG:%f INCR_X:%f INCR_Y;%f\n",rb.incr_ang*180.0/M_PI,rb.incr_x,rb.incr_y);
    rb.PublishOdom6d();
    Move();
/*
    if (msg.joy_butt_state[DS_CONTR_MAN_AUTO]==AUTO_CONTROL){
      //Send back references to the GPRMC node
      rb.TwistToTracks(rb.ref_joy_x,rb.ref_joy_y,&ref1_joy,&ref2_joy);
      printf("LOCAL_JOY_CONTROL: RefX:%f RefY:%f Axis1_ref:%d Axis2_ref:%d\n",rb.ref_joy_x,rb.ref_joy_y,ref1_joy,ref2_joy);

      //printf("EUSEBIO:Publishing references\n");
      rb.CtrlRobot(ref1_joy,ref2_joy,rb.robot_st.dig_ctr_state);
    }
*/
/*
    //-------------- ROBOT STATUS UPDATE -----------------
    rb.robot_st.header.stamp=msg.header.stamp;
    rb.robot_st.hour=msg.hour;
    rb.robot_st.min=msg.min;
    rb.robot_st.sec=msg.sec;

    //Battery status
    rb.robot_st.battery_level=msg.volt_values[0];
    //printf("Battery:%f volts\n",msg.volt_values[0]);

    //Digital IO status
    rb.robot_st.dig_robot_state.clear();
    for (int i=0; i< (int)msg.joy_butt_state.size();i++){
      rb.robot_st.dig_robot_state.push_back(msg.joy_butt_state[i]);
    }

    //Cinematic status
    if (rb.incr_t) rb.robot_st.linear_speed+=K_SPEED_COEF*(rb.incr_dist_robot/rb.incr_t-rb.robot_st.linear_speed);
    //if (rb.incr_t) rb.robot_st.linear_speed=(rb.incr_dist_robot/rb.incr_t);
    else rb.robot_st.linear_speed=0.0;

    //Control status
    rb.robot_st.ref_steer=(float)rb.ref_joy_x;
    rb.robot_st.ref_vel=(float)rb.ref_joy_y;

    //printf("Speed: %f (m/s)\n",rb.robot_st.linear_speed);
    rb.robot_st.total_distance_travelled+=fabs(rb.incr_dist_robot);
    rb.robot_st.total_working_time+=rb.incr_t;

    //printf("Working time:%f (secs)\n",rb.robot_st.total_working_time);
    rb.robot_st.partial_distance_travelled+=fabs(rb.incr_dist_robot);
    rb.robot_st.partial_working_time+=rb.incr_t;
    if (rb.incr_dist_robot && (msg.joy_y1 || msg.joy_x1 || rb.ref_joy_x || rb.ref_joy_y)){
      rb.robot_st.total_moving_time+=rb.incr_t;
      rb.robot_st.partial_moving_time+=rb.incr_t;
    }

    //printf("PUBLISHING ROBOT STATUS....................\n");
    rb.pub_robot_status.publish(rb.robot_st);
    rb.WriteRobotStatus(rb.robot_st, robot_status_filename);
*/
}


int main(int argc, char **argv)
{
  for (int i=1;i<argc;i++){
    if (!strcmp(argv[i],"-f")){//Load robot parameters file in YAML format
       rb.ReadConfigurationFile((argv[i+1]));
    }else if (!strcmp(argv[i],"-status")){//Load robot parameters file in YAML format
        robot_status_filename.assign(argv[i+1]);
        rb.ReadRobotStatus(&rb.robot_st,robot_status_filename);
    }else if (!strcmp(argv[i],"-ch_pose")){//Read robot charging pose file in YAML format
        charging_pose_filename.assign(argv[i+1]);
        rb.ReadChargingPose(charging_pose_filename);
      }
  }
 // rb.robot_st.linear_speed=0.0;
  //rb.robot_st.steering_ang=0.0;

  ros::init(argc, argv, "eusebio");
  ros::NodeHandle nh;

  //Publishing
  rb.pub_gpmrc_command = nh.advertise<gpmrc_node::gpmrc_command>("/gpmrc/command", 1);
  rb.pub_robot_status = nh.advertise<eusebio::RobotStatus>("/eusebio/RobotStatus", 1);
  rb.pub_marker = nh.advertise<visualization_msgs::Marker>("/eusebio/markers", 1);
  rb.pub_ch_pose = nh.advertise< geometry_msgs::PoseWithCovarianceStamped> ("/initialpose", 1);


  //Subscribing
  //ros::Subscriber mov_axes_sub=nh.subscribe<joy::Joy> ("/joy", 1, JoyHnd);
  ros::Subscriber mov_axes_sub=nh.subscribe<sensor_msgs::Joy> ("/joy", 1, JoyHnd);
  ros::Subscriber gpmrc_state_sub=nh.subscribe<gpmrc_node::gpmrc_state>("/gpmrc/state",1000,GPMRCStateHnd);

  ROS_INFO("Eusebio robot node ready...");
  tf::TransformListener listener1,listener2,listener3;

  listener1.waitForTransform(rb.map_frame_name,rb.odom_frame_name,ros::Time::now(),ros::Duration(1.0));
  //listener2.waitForTransform(rb.odom_frame_name,rb.base_frame_name,ros::Time::now(),ros::Duration(1.0));
  //listener3.waitForTransform(rb.base_frame_name,rb.base_center_name,ros::Time::now(),ros::Duration(1.0));

  //Main loop
  //ros::spin();
  ros::Rate rate(rb.fps);
  while (ros::ok()){
    try{//Catch amcl map corrections
      double qx,qy,qz,qw,x,y,z;
      double rot[9];
      tf::StampedTransform transform;

/*
      rb.mt.ResetTransform(&rb.correction_transf[0]);
      listener1.lookupTransform(rb.map_frame_name,rb.odom_frame_name,ros::Time(0),transform);

      x=transform.getOrigin().x();
      y=transform.getOrigin().y();
      z=transform.getOrigin().z();

      qx=(double)transform.getRotation().x();
      qy=(double)transform.getRotation().y();
      qz=(double)transform.getRotation().z();
      qw=(double)transform.getRotation().w();

      rb.mt.FromQuaternionToRotationMatrix(qx,qy,qz,qw,&rot[0]);

      rb.mt.setRotMatrixInTransf(&rot[0],&rb.correction_transf[0]);

      rb.correction_transf[3]=x;
      rb.correction_transf[7]=y;
      rb.correction_transf[11]=z;
*/
    }catch(tf::TransformException ex){
      ROS_ERROR("%s",ex.what());
    }

    ros::spinOnce();
    rate.sleep();
  }
}



