#include "robot.h"


void Robot::CtrlRobot(int16_t reflin, int16_t refang, std::vector<unsigned char> buttons){
  gpmrc_node::gpmrc_command cmd;

  cmd.vel_left=-(int)(reflin-refang);
  cmd.vel_right=-(int)(reflin+refang);

  this->pub_gpmrc_command.publish(cmd);
}

void Robot::TwistToTracks(double ref_x,double ref_y,int16_t *ref1,int16_t *ref2){
    int ref_ang,ref_lin,ref1_joy=0,ref2_joy=0;
    double radius,alpha;

    radius=sqrt(pow(ref_x,2.0)+pow(ref_y,2.0));

    if (radius>VEL_OFFSET) alpha=atan2(ref_x,ref_y);
    else alpha=0.0;

    //printf("Radius:%f Alpha:%f\n",radius,alpha);

    ref_lin=(int)(-radius*cos(alpha)*VEL_JOY_LIN_COEF);

    if (alpha>(M_PI/2.0+VEL_OFFSET)){
      alpha=-M_PI+alpha;
    }else if (alpha<-(M_PI/2.0+VEL_OFFSET)){
      alpha=M_PI+alpha;
    }

    ref_ang=(int)(radius*alpha*VEL_JOY_ANG_COEF);
  *ref1=ref_x*(this->max_vel_pls);//Experimental
  *ref2=ref_y*(this->max_vel_pls);//Experimental
}

void Robot::PublishOdom6d(void)
{
  double qx,qy,qz,qw;
  double tstamp_now;
  static double last_tstamp=0.0;

  tstamp_now=(double)this->timestamp.sec+(double)this->timestamp.nsec*1E-9;

  if (last_tstamp){
    static tf::TransformBroadcaster br;
    geometry_msgs::TransformStamped out_transform;


    //Odometry transform publication
    out_transform.header.stamp=this->timestamp;
    out_transform.header.frame_id=this->odom_frame_name;
    out_transform.child_frame_id=this->base_frame_name;

    out_transform.transform.translation.x=this->xs_odom;
    out_transform.transform.translation.y=this->ys_odom;
    out_transform.transform.translation.z=0.0;

    this->mt.FromEulerToQuaternion(0.0, 0.0, this->theta_odom,&qx,&qy,&qz,&qw);

    out_transform.transform.rotation.w=qw;
    out_transform.transform.rotation.x=qx;
    out_transform.transform.rotation.y=qy;
    out_transform.transform.rotation.z=qz;

    br.sendTransform(out_transform);

    //FRAMES TF
    this->mt.FromEulerToQuaternion(0.0,0.0,this->laser_angle_pos,&qx,&qy,&qz,&qw );

    out_transform.header.stamp=this->timestamp;
    out_transform.header.frame_id=this->base_frame_name;
    out_transform.child_frame_id=this->laser_frame_name;

    //printf("Base frame name:%s Laser frame name:%s\n",this->base_frame_name.c_str(),this->laser_frame_name.c_str());
    out_transform.transform.translation.x=this->laser_long_pos;
    out_transform.transform.translation.y=0;
    out_transform.transform.translation.z=this->laser_height_pos;
    out_transform.transform.rotation.w=qw;
    out_transform.transform.rotation.x=qx;
    out_transform.transform.rotation.y=qy;
    out_transform.transform.rotation.z=qz;

    br.sendTransform(out_transform);

/*
    out_transform.header.stamp=this->timestamp;
    out_transform.header.frame_id=this->base_frame_name;
    out_transform.child_frame_id=this->wii_ir_frame_name;

    out_transform.transform.translation.x=this->wii_ir_long_pos;
    out_transform.transform.translation.y=0;
    out_transform.transform.translation.z=this->wii_ir_height;
    out_transform.transform.rotation.w=1.0;
    out_transform.transform.rotation.x=0.0;
    out_transform.transform.rotation.y=0.0;
    out_transform.transform.rotation.z=0.0;

    br.sendTransform(out_transform);
    */
  }
  last_tstamp=tstamp_now;

 // this->PublishPosition();

}



void Robot::ReadConfigurationFile(char *filename){
    std::ifstream conf_file(filename);
    YAML::Parser parser(conf_file);
    YAML::Node doc;

    printf("Reading target map: %s...\n",filename);
    while(parser.GetNextDocument(doc)){
        if (const YAML::Node *pVal = doc.FindValue("RosFPS")){
          *pVal >> this->fps;
          printf("RosFPS:%f\n",this->fps);
        }
        if (const YAML::Node *pVal = doc.FindValue("PlsDist")){
          *pVal >> this->pls_dist;
          printf("PlsDist:%f\n",this->pls_dist);
        }
        if (const YAML::Node *pVal = doc.FindValue("LaserLongPos")){
          *pVal >> this->laser_long_pos;
          printf("LaserLongPos:%f\n",this->laser_long_pos);
        }
        if (const YAML::Node *pVal = doc.FindValue("LaserHeightPos")){
          *pVal >> this->laser_height_pos;
          printf("LaserHeightPos:%f\n",this->laser_height_pos);
        }
        if (const YAML::Node *pVal = doc.FindValue("LaserAnglePos")){
          *pVal >> this->laser_angle_pos;
          printf("LaserAnglePos(grad):%f\n",this->laser_angle_pos);
          this->laser_angle_pos*=M_PI/180.0;
        }
        if (const YAML::Node *pVal = doc.FindValue("DistBetweenWheels")){
          *pVal >> this->dist_between_wheels;
          printf("DistBetweenWheels:%f\n",this->dist_between_wheels);
        }
        if (const YAML::Node *pVal = doc.FindValue("WiiIrLongPos")){
          *pVal >> this->wii_ir_long_pos;
          printf("WiiIrLongPos:%f\n",this->wii_ir_long_pos);
        }
        if (const YAML::Node *pVal = doc.FindValue("WiiIrHeight")){
          *pVal >> this->wii_ir_height;
          printf("WiiIrHeight:%f\n",this->wii_ir_height);
        }
        if (const YAML::Node *pVal = doc.FindValue("MaxVelPls")){
          *pVal >> this->max_vel_pls;
          printf("MaxVelPls:%f\n",this->max_vel_pls);
        }
        if (const YAML::Node *pVal = doc.FindValue("OdomFrame")){
          *pVal >> this->odom_frame_name;
          printf("OdomFrame:%s\n",this->odom_frame_name.c_str());
        }
        if (const YAML::Node *pVal = doc.FindValue("MapFrame")){
          *pVal >> this->map_frame_name;
          printf("MapFrame:%s\n",this->map_frame_name.c_str());
        }
        if (const YAML::Node *pVal = doc.FindValue("BaseFrame")){
          *pVal >> this->base_frame_name;
          printf("BaseFrame:%s\n",this->base_frame_name.c_str());
        }
        if (const YAML::Node *pVal = doc.FindValue("LaserFrame")){
          *pVal >> this->laser_frame_name;
          printf("LaserFrame:%s\n",this->laser_frame_name.c_str());
        }
        if (const YAML::Node *pVal = doc.FindValue("BaserRearFrame")){
          *pVal >> this->base_rear_axis_name;
          printf("BaseRearFrame:%s\n",this->base_rear_axis_name.c_str());
        }
        if (const YAML::Node *pVal = doc.FindValue("BaseCenterFrame")){
          *pVal >> this->base_center_name;
          printf("BaseCenterFrame:%s\n",this->base_center_name.c_str());
        }
        if (const YAML::Node *pVal = doc.FindValue("WiiIRFrame")){
          *pVal >> this->wii_ir_frame_name;
          printf("WiiIRFrame:%s\n",this->wii_ir_frame_name.c_str());
        }
        if (const YAML::Node *pVal = doc.FindValue("RobotLength")){
          *pVal >> this->robot_length;
          printf("RobotLength:%f\n",this->robot_length);
        }
        if (const YAML::Node *pVal = doc.FindValue("RobotWidth")){
          *pVal >> this->robot_width;
          printf("RobotWidth:%f\n",this->robot_width);
        }
        if (const YAML::Node *pVal = doc.FindValue("RobotHeight")){
          *pVal >> this->robot_height;
          printf("RobotHeight:%f\n",this->robot_height);
        }
    }
}

void operator >> (const YAML::Node &node, eusebio::RobotStatus &rs){
    node["TotalDistanceTravelled"]>>rs.total_distance_travelled;
    node["TotalWorkingTime"]>>rs.total_working_time;
    node["TotalMovingTime"]>>rs.total_moving_time;
}

void Robot::ReadRobotStatus(eusebio::RobotStatus *rs,string filename){
    std::ifstream conf_file(filename.c_str());
    YAML::Parser parser(conf_file);
    YAML::Node doc;

    printf("Reading robot status file: %s...\n",filename.c_str());
    while(parser.GetNextDocument(doc)){
        if (const YAML::Node *pVal = doc.FindValue("TotalDistanceTravelled")){
          *pVal >> rs->total_distance_travelled;
          printf("TotalDistanceTravelled:%f\n",rs->total_distance_travelled);
        }
        if (const YAML::Node *pVal = doc.FindValue("TotalMovingTime")){
            *pVal >> rs->total_moving_time;
          printf("TotalMovingTime:%f\n",rs->total_moving_time);
        }
        if (const YAML::Node *pVal = doc.FindValue("TotalWorkingTime")){
          *pVal >> rs->total_working_time;
          printf("TotalWorkingTime:%f\n",rs->total_working_time);
        }
    }
}

void Robot::WriteRobotStatus(eusebio::RobotStatus rs, string filename){
  YAML::Emitter out;

  out << YAML::BeginMap << YAML::Key << "TotalDistanceTravelled";
  out << YAML::Value << rs.total_distance_travelled;
  out << YAML::Key << "TotalMovingTime";
  out << YAML::Value << rs.total_moving_time;
  out << YAML::Key << "TotalWorkingTime";
  out << YAML::Value << rs.total_working_time;

  out << YAML::EndMap;

  std::ofstream fout(filename.c_str());
  fout << out.c_str();
}

void Robot::SetChargingPose(void){
  geometry_msgs::PoseWithCovarianceStamped msg;

  printf("Setting Charging Pose: X:%f Y:%f QX:%f QY:%f QZ:%f QW:%f\n",this->charging_pose_x, this->charging_pose_y,this->charging_pose_qx,this->charging_pose_qy,this->charging_pose_qz,this->charging_pose_qw);
  msg.pose.pose.position.x=this->charging_pose_x;
  msg.pose.pose.position.y=this->charging_pose_y;
  msg.pose.pose.position.z=0.0;
  msg.pose.pose.orientation.x=this->charging_pose_qx;
  msg.pose.pose.orientation.y=this->charging_pose_qy;
  msg.pose.pose.orientation.z=this->charging_pose_qz;
  msg.pose.pose.orientation.w=this->charging_pose_qw;
  this->pub_ch_pose.publish(msg);
}

void Robot::ReadChargingPose(string filename){
    std::ifstream conf_file(filename.c_str());
    YAML::Parser parser(conf_file);
    YAML::Node doc;

    printf("Reading charging pose file: %s...\n",filename.c_str());
    while(parser.GetNextDocument(doc)){
        if (const YAML::Node *pVal = doc.FindValue("X")){
          *pVal >> this->charging_pose_x;
          printf("Charging pose X:%f\n",this->charging_pose_x);
        }
        if (const YAML::Node *pVal = doc.FindValue("Y")){
          *pVal >> this->charging_pose_y;
          printf("Charging pose Y:%f\n",this->charging_pose_y);
        }
        if (const YAML::Node *pVal = doc.FindValue("QX")){
          *pVal >> this->charging_pose_qx;
          printf("Charging pose QX:%f\n",this->charging_pose_qx);
        }
        if (const YAML::Node *pVal = doc.FindValue("QY")){
          *pVal >> this->charging_pose_qy;
          printf("Charging pose QY:%f\n",this->charging_pose_qy);
        }
        if (const YAML::Node *pVal = doc.FindValue("QZ")){
          *pVal >> this->charging_pose_qz;
          printf("Charging pose QZ:%f\n",this->charging_pose_qz);
        }
        if (const YAML::Node *pVal = doc.FindValue("QW")){
          *pVal >> this->charging_pose_qw;
          printf("Charging pose QW:%f\n",this->charging_pose_qw);
        }
    }
}

void Robot::PublishRobotMarker(void){
    visualization_msgs::Marker marker_robot;
    //double qx,qy,qz,qw;

    marker_robot.header.frame_id = "/map";
    marker_robot.header.stamp = ros::Time::now();
    marker_robot.ns = "robot";
    marker_robot.id = 0;
    marker_robot.type = visualization_msgs::Marker::CUBE;

    marker_robot.action = visualization_msgs::Marker::ADD;

    marker_robot.pose.position.x = this->robot_st.xc;
    marker_robot.pose.position.y = this->robot_st.yc;
    marker_robot.pose.position.z = 0.0;

    //this->mt.FromEulerToQuaternion(0.0, 0.0, this->roll,&qz,&qy,&qz,&qw);

    marker_robot.pose.orientation.x = this->qx;
    marker_robot.pose.orientation.y = this->qy;
    marker_robot.pose.orientation.z = this->qz;
    marker_robot.pose.orientation.w = this->qw;

    marker_robot.scale.x = this->robot_length;
    marker_robot.scale.y = this->robot_width;
    marker_robot.scale.z = this->robot_height;

    marker_robot.color.r = 1.0f;
    marker_robot.color.g = 1.0f;
    marker_robot.color.b = 1.0f;
    marker_robot.color.a = 0.5;
    marker_robot.lifetime = ros::Duration();
    this->pub_marker.publish(marker_robot);
}

void Robot::PublishChargerMarker(unsigned int status){
    visualization_msgs::Marker marker;

    marker.header.frame_id = this->mark_frame_name;
    marker.header.stamp = ros::Time::now();
    marker.ns = "robot_status";
    if (status==(unsigned int) ROBOT_CHARGING){
      marker.id = 0;
      marker.type = visualization_msgs::Marker::CUBE;

      marker.action = visualization_msgs::Marker::ADD;

      marker.pose.position.x = 0.0;
      marker.pose.position.y = 0.0;
      marker.pose.position.z = 0.0;

      marker.pose.orientation.x = 0.0;
      marker.pose.orientation.y = 0.0;
      marker.pose.orientation.z = 0.0;
      marker.pose.orientation.w = 1.0;

      marker.scale.x = 1.0;
      marker.scale.y = 1.0;
      marker.scale.z = 1.0;

      marker.color.r = 0.0f;
      marker.color.g = 1.0f;
      marker.color.b = 0.0f;
      marker.color.a = 0.5;
      marker.lifetime = ros::Duration();
      this->pub_marker.publish(marker);
  }else if (status==(unsigned int) ROBOT_CONNECTED_NOT_CHARGING){
      marker.id = 0;
      marker.type = visualization_msgs::Marker::CUBE;

      marker.action = visualization_msgs::Marker::ADD;

      marker.pose.position.x = 0.0;
      marker.pose.position.y = 0.0;
      marker.pose.position.z = 0.0;

      marker.pose.orientation.x = 0.0;
      marker.pose.orientation.y = 0.0;
      marker.pose.orientation.z = 0.0;
      marker.pose.orientation.w = 1.0;

      marker.scale.x = 1.0;
      marker.scale.y = 1.0;
      marker.scale.z = 1.0;

      marker.color.r = 1.0f;
      marker.color.g = 0.0f;
      marker.color.b = 0.0f;
      marker.color.a = 0.5;
      marker.lifetime = ros::Duration();
      this->pub_marker.publish(marker);
   }else if (status==(unsigned int) ROBOT_NOT_CONNECTED){
      marker.id = 0;
      marker.type = visualization_msgs::Marker::CUBE;

      marker.action = visualization_msgs::Marker::ADD;

      marker.pose.position.x = 0.0;
      marker.pose.position.y = 0.0;
      marker.pose.position.z = 0.0;

      marker.pose.orientation.x = 0.0;
      marker.pose.orientation.y = 0.0;
      marker.pose.orientation.z = 0.0;
      marker.pose.orientation.w = 1.0;

      marker.scale.x = 1.0;
      marker.scale.y = 1.0;
      marker.scale.z = 1.0;

      marker.color.r = 0.0f;
      marker.color.g = 0.0f;
      marker.color.b = 1.0f;
      marker.color.a = 0.5;
      marker.lifetime = ros::Duration();
      this->pub_marker.publish(marker);
   }
}

