/***************************************************************************
 *   Copyright (C) 2008 by JAime Alberto Pulido Fentanes   *
 *   jaipul@cartif.es   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "serial_port/lightweightserial.h"

#include "ros/ros.h"
#include "std_msgs/String.h"

#include <boost/thread.hpp>
#include <iostream>

#include "gpmrc_node/gpmrc_state.h"
#include "gpmrc_node/AxisData.h"
#include "gpmrc_node/gpmrc_command.h"

#define MAX_SUBCADENAS	50
//#define MAX_SAVE_COUNT	10
#define MAXBUFF 512
#define MAX_INCR_ODOM 2000

#define NUM_FUNCTIONALITIES 7

gpmrc_node::gpmrc_state ros_gpmrc_state;
ros::Publisher pub_gpmrc_state;
LightweightSerial *serial_port=0;

//void sig_IO (int status);
void ReceiveSerial(void);
void ReceiveROS(void);
int ReadData(char *ucData);
int ReadByte(char *byte);
void SendSerialCommand(char *command, unsigned int Lcommand);
bool Decode(char * Info, unsigned int nbytes);
int ExtractArguments(char *cad,char *array[],const char *separator);
void CommandHnd(gpmrc_node::gpmrc_command msg);
void ID_RequestHnd (void);
void Publish(void);

void CommandHnd(gpmrc_node::gpmrc_command msg){
  char WriteBuff[MAXBUFF];
  unsigned int length;

  sprintf(WriteBuff,"$REF:[3,%d,%d]#\r\n",msg.vel_left,msg.vel_right);
  length=(unsigned int)strlen(WriteBuff);

  printf("SENDING: (%d) %s\n",length,WriteBuff);
  SendSerialCommand((char *)WriteBuff, length);
}

void ReceiveSerial(void){
  int Bytesread=0;
  char Readbuff[MAXBUFF];

  printf("THREAD_SERIAL: LISTENING ON THE SERIAL PORT\n");

  while(1){
    if (!(Bytesread=ReadData(Readbuff))){
      printf("Read data Error\n");
    }else{
      //printf("[%d] %s\n",Bytesread,Readbuff);

      if (Decode(Readbuff,Bytesread)==true){//Decodifica la informacin de las tramas y rellena la estructura de estado          
        Publish();
      }
    }
  }
}

void ReceiveROS(void){
    printf("THREAD_ROS: LISTENING ROS TOPICS\n");
    ros::Rate rate(50);
    while (ros::ok()){
      ros::spinOnce();
      rate.sleep();
    }
}

void Publish(void){
    time_t t;
    struct tm *tm;
    static double t_ant=0.0;
    double t_now,t_lapsed;

    //printf("PUBLISHING GPMRC STATE\n");

    time(&t);
    tm=localtime(&t);

    ros_gpmrc_state.header.stamp=ros::Time::now();

    ros_gpmrc_state.hour=(int32_t)tm->tm_hour;
    ros_gpmrc_state.min=(int32_t)tm->tm_min;
    ros_gpmrc_state.sec=(int32_t)tm->tm_sec;
    t_now=(double)(ros_gpmrc_state.header.stamp.sec+ros_gpmrc_state.header.stamp.nsec*1E-9);
    t_lapsed=t_now-t_ant;
    t_ant=t_now;
    //printf("LAPSED TIME:%f\n",t_lapsed);

    pub_gpmrc_state.publish(ros_gpmrc_state);

    ros_gpmrc_state.volt_read=false;
    ros_gpmrc_state.joy_read=false;
    ros_gpmrc_state.mot_read=false;
    ros_gpmrc_state.son_read=false;

    //GuardaEstadoEnXML(xmlfile);//Guarda la estructura de estado en fichero XML
}


void ID_RequestHnd (void)
{
  char data[]="$ID_:#\r\n";
  unsigned int length;

  length=(unsigned int)strlen(data);
  printf("SENDING IDENTITY REQUEST TO THE CONTROLLER\n[%d] %s",length,data);
  SendSerialCommand(data,length);
}

int ReadData(char *cData){
    char cCod=' ';

    //Check header ($)
    //printf("Searching header $\n");
    do {
      if(ReadByte(&cCod)<=0){
        return 0;
      }
    }while(cCod!='$');
    cData[0]=cCod;

    //Check termination
    //printf("Searching terminator #\n");
    int i=1;
    do{
      if (ReadByte(&cCod)<=0){
        return 0;
      }
      if (cCod!='\n') {
        cData[i]=cCod;
        i++;
      }
    }while((cData[i-1]!='#') && (i<MAXBUFF));
    if (cData[i-1]=='#'){//Terminacin de trama
     //printf("Searching CR and LF\n");
     if (ReadByte(&cCod)<=0){//\r
       //fprintf(stderr,"No se pudo leer cabecera (0xFF)\n");
       printf("No se pudo leer byte\n");
       return 0;
     }
     if (ReadByte(&cCod)<=0){//\n
        //fprintf(stderr,"No se pudo leer cabecera (0xFF)\n");
        printf("No se pudo leer byte\n");
        return 0;
     }
     //printf("End of msg. Read\n");
    }else printf("End of buffer reached\n");
    cData[i]='\0';
    return i;
}

void SendSerialCommand(char *command, unsigned int Lcommand){
  int iBytesWritten;

  //printf("SENDING: %s\n",command);

  if (serial_port->happy==true){
    iBytesWritten=serial_port->serial_writen(command, Lcommand);
    //printf("Lcommand:%d Written:%d\n",(int)Lcommand,iBytesWritten);
    if (iBytesWritten!=(int)Lcommand) {
      fprintf(stderr,"Couldn't send command to the controller\n");
      return;
    }//else printf("Command sent\n");
  }else printf("Serial port is not happy\n");

}

int ReadByte(char *byte){
  int iNumRead;

  iNumRead=serial_port->serial_readn(byte,1);
  switch(iNumRead)  {
    case -1:
      printf("Port not openned\n");
      return 0;
      break;
    case -2:
      printf("Timeout error\n");
      return 0;
      break;
    case 1:
      //printf("%c\n",byte[0]);
      return 1;
      break;
    default:
      printf("Couldn't read byte from seria port \n");
      return 0;
      break;
  }
  return 0;
}

bool Decode(char * Info, unsigned int nbytes){
  char *dat_args[MAX_SUBCADENAS],*func_str=NULL,*arg_str=NULL;
  int num_dat,i,j,k=0;


  //printf("STEP1: [%d] %s\n",strlen(Info),Info);
  if(Info[0] == '$'){//Header byte '$'
     func_str=(char *)malloc((strlen(Info)+1)*sizeof(char));

     //Extract the information between $ and #
     j=0;
     for(i=0;i<(int)strlen(Info);i++){
       if ((Info[i]!='$')&&(Info[i]!='#')&&(Info[i]!='\n')&&(Info[i]!='\r')){
          func_str[j]=Info[i];
          j++;
       }
     }
     func_str[j]='\0';
     //printf("STEP2: [%d]  %s\n",strlen(func_str),func_str);

     //Extract string between []
     arg_str=(char *)malloc((strlen(func_str)+1)*sizeof(char));
     for(i=0;i<(int)strlen(func_str);i++){
        if (func_str[i]=='[') {
          k=i+1;
          break;
        }
      }

      j=0;
      for(i=k;i<(int)strlen(func_str);i++){
        if (func_str[i]!=']') {
           arg_str[j]=func_str[i];
           j++;
        }else break;
      }
      arg_str[j]='\0';
      //printf("STEP3: [%d]  %s\n",strlen(arg_str),arg_str);

      //Extract arguments separated by ','
      if ((num_dat=ExtractArguments(arg_str,dat_args,","))<0){
        if(!strncmp("ACK:",func_str,4)) {
           return true;
        }else return false;
      }
      //printf("NUM_ARGS:%d\n",num_dat);

      if(!strncmp("ID_:",func_str,4)) {
         int numf,q;

         //if (strcmp(dat_args[0],"GPMRC6_EUSEBIO")){
         //if (strcmp(dat_args[0],"GPMRC6_AGV_2")){
         //   printf("WRONG FIRMWARE IDENTIFIER (%s). THIS NODE IS DESIGNED FOR GPMRC6_AGV\n",dat_args[0]);
         //   return false;
         //}else{
             printf("FIRMWARE VERSION:%s\n",dat_args[0]);
            ros_gpmrc_state.contr_name=dat_args[0];
         //}

         numf=atoi(dat_args[1]);
         printf("NUMBER OF FUNCIONALITIES:%d\n",numf);
         ros_gpmrc_state.num_funcs=(int8_t)numf;
         if (numf!=num_dat-2) {
            printf("FUNCTIONALITIES STRING MISSMATCH\n");
            return false;
         }

         ros_gpmrc_state.func_str.clear();

         for(q=0;q<numf;q++){
             printf("FUNCIONALITY:%s ACTIVE\n",dat_args[q+2]);
             if (!strcmp(dat_args[q+2],"VLT")){//Voltage Control
               ros_gpmrc_state.volt_active=true;
               ros_gpmrc_state.func_str.push_back("VLT");
             }else if (!strcmp(dat_args[q+2],"JOY")){//Joystick
               ros_gpmrc_state.joy_active=true;
               ros_gpmrc_state.func_str.push_back("JOY");
             }else if (!strcmp(dat_args[q+2],"MOT")){//Motor Control
               ros_gpmrc_state.mot_active=true;
               ros_gpmrc_state.func_str.push_back("MOT");
             }else if (!strcmp(dat_args[q+2],"PC1")){//RS232-1 Comunicacions
               ros_gpmrc_state.com1_active=true;
               ros_gpmrc_state.func_str.push_back("PC1");
             }else if (!strcmp(dat_args[q+2],"PC2")){//RS232-2 Comunicacions
               ros_gpmrc_state.com2_active=true;
               ros_gpmrc_state.func_str.push_back("PC2");
             }else if (!strcmp(dat_args[q+2],"AL_")){//Alarm Control
               ros_gpmrc_state.alrm_active=true;
               ros_gpmrc_state.func_str.push_back("AL_");
             }else if (!strcmp(dat_args[q+2],"LCD")){//LCD visualization msgs
               ros_gpmrc_state.LCD_active=true;
               ros_gpmrc_state.func_str.push_back("LCD");
             }else if (!strcmp(dat_args[q+2],"CAN")){//CAN bus Comunications
             }else if (!strcmp(dat_args[q+2],"ACT")){//Linear Actuator
               ros_gpmrc_state.lin_act_active=true;
               ros_gpmrc_state.func_str.push_back("ACT");
             }else if (!strcmp(dat_args[q+2],"ITM")){//Blinking light
               ros_gpmrc_state.bl_active=true;
               ros_gpmrc_state.func_str.push_back("ITM");
             }else if (!strcmp(dat_args[q+2],"SON")){//Lectura sonars
               ros_gpmrc_state.son_active=true;
               ros_gpmrc_state.func_str.push_back("SON");
             }else if (!strcmp(dat_args[q+2],"BMP")){//Lectura bumpers
               ros_gpmrc_state.bmp_active=true;
               ros_gpmrc_state.func_str.push_back("BMP");
             }else if (!strcmp(dat_args[q+2],"REL")){//Activacion/Desactivacion reles
               ros_gpmrc_state.rel_active=true;
               ros_gpmrc_state.func_str.push_back("REL");
             }

          }
      }else if(!strncmp("VLT:",func_str,4)) {//Voltages
         int q,num;

         ros_gpmrc_state.volt_active=true;
         ros_gpmrc_state.volt_read=true;

         num=atoi(dat_args[0]);//Number of voltages
         if (num!=num_dat-1){
            printf("WRONG VOLTAGE MESSAGE.........................................................................................\n");
            return false;
         }

         ros_gpmrc_state.volt_num=(int8_t)num;

         ros_gpmrc_state.volt_values.clear();

         for(q=0;q<num;q++){
            float vlt;

            sscanf(dat_args[q+1],"%f",&vlt);
            ros_gpmrc_state.volt_values.push_back((float)vlt);
         }
          //printf("Volt:%f\n",ros_gpmrc_state.volt_values[0]);

         //printf("]\n");
         //printf("VLT-%s\t",arg_str);
      }else if(!strncmp("JOY:",func_str,4)) {//Joystick
          int nbutt;

          ros_gpmrc_state.joy_active=true;
          ros_gpmrc_state.joy_read=true;

          ros_gpmrc_state.joy_src=(int8_t)atoi(dat_args[0]);
          ros_gpmrc_state.joy_x1=(int16_t)atoi(dat_args[1]);
          ros_gpmrc_state.joy_y1=(int16_t)atoi(dat_args[2]);
          ros_gpmrc_state.joy_x2=(int16_t)atoi(dat_args[3]);
          ros_gpmrc_state.joy_y2=(int16_t)atoi(dat_args[4]);

          //printf("JOYSTICK2:X=%s Y=%s\n",dat_args[3],dat_args[4]);

          nbutt=atoi(dat_args[5]);//N de botones del joystick

          if (nbutt!=num_dat-6){
             printf("WRONG JOYSTICK MESSAGE\n");
             return false;
          }

          ros_gpmrc_state.joy_nbutt=(int8_t)nbutt;

          ros_gpmrc_state.joy_butt_state.clear();
          for (i=0;i<nbutt;i++){
            ros_gpmrc_state.joy_butt_state.push_back((bool)atoi(dat_args[6+i]));
          }
          //printf("]\n");
       }else if(!strncmp("MOT:",func_str,4)) {//Motors
         int nmot;
         double dat_aux;
         gpmrc_node::AxisData ad;

         ros_gpmrc_state.mot_active=true;
         ros_gpmrc_state.mot_read=true;

         nmot=atoi(dat_args[0]);//Number of motors

         ros_gpmrc_state.mot_num=(int8_t)nmot;

         if (ros_gpmrc_state.axes.size()!=nmot) {
           ros_gpmrc_state.axes.clear();
           ad.odo=0;
           ad.vel=0;
           ad.cal=true;
           for (unsigned int i=0;i<nmot;i++) ros_gpmrc_state.axes.push_back(ad);
         }

         for (int i=0;i<nmot;i++){
           //ros_gpmrc_state.axes.clear();

           dat_aux=(double)atoi(dat_args[2*i+1]);//Motor speed
           ad.vel=(int32_t)(dat_aux);

           dat_aux=(double)atoi(dat_args[2*i+2]);//Motor odometry
           ad.odo=(int32_t)(dat_aux);
           if ((ad.vel<=0)&&(ad.odo-ros_gpmrc_state.axes[i].odo)>MAX_INCR_ODOM){
             //printf("ODOMETRY JUMP DETECTED. AXIS:%d\n",i+1);
             ad.vel=ad.odo-ros_gpmrc_state.axes[i].odo-65535;
             //ad.vel=0;
           }else if ((ad.vel>=0)&&(ad.odo-ros_gpmrc_state.axes[i].odo)<-MAX_INCR_ODOM){
             //printf("ODOMETRY JUMP DETECTED. AXIS:%d\n",i+1);
             ad.vel=65535-(ros_gpmrc_state.axes[i].odo-ad.odo);
             //ad.vel=0;
           }else {
             ad.vel=(int32_t)(ad.odo-ros_gpmrc_state.axes[i].odo);
           }
           ros_gpmrc_state.axes[i].odo=ad.odo;
           ros_gpmrc_state.axes[i]=ad;
         }
        //printf("]\n");
      }else if(!strncmp("AL_:",func_str,4)) {//Alarms
         char str_alrm[30];

         ros_gpmrc_state.alrm_active=true;

         //printf("ALARM STRING:%s\n",dat_args[0]);

         strncpy(str_alrm,(char *)&dat_args[0][0],1);
         ros_gpmrc_state.alrm_lv=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][1],1);
         ros_gpmrc_state.alrm_vlv=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][2],1);
         ros_gpmrc_state.alrm_emg=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][3],1);
         ros_gpmrc_state.alrm_joy_com=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][4],1);
         ros_gpmrc_state.alrm_enc_com=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][5],1);
         ros_gpmrc_state.alrm_pot_com=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][6],1);
         ros_gpmrc_state.alrm_aux2=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][7],1);
         ros_gpmrc_state.alrm_aux1=(bool)atoi(str_alrm);

         strncpy(str_alrm,(char *)&dat_args[0][8],1);
         ros_gpmrc_state.alrm_rst=(bool)atoi(str_alrm);

         //Publish();
      }else if(!strncmp("ACT:",func_str,4)) {//Lineal actuator state
          ros_gpmrc_state.lin_act_active=true;
          ros_gpmrc_state.lin_act_read=true;
          ros_gpmrc_state.lin_act_state=(int)atoi(dat_args[0]);
          //printf("ACT (Linear_actuator): State:%d\n",(int)ros_gpmrc_state.lin_act_state);
      }else if(!strncmp("ITM:",func_str,4)) {//Blinking lights state
         ros_gpmrc_state.bl_active=true;
         ros_gpmrc_state.bl_read=true;
         ros_gpmrc_state.bl_state=(bool)atoi(dat_args[0]);
         //printf("ITM (Blinking ligths): State:%d [",(int)ros_gpmrc_state.bl_state);
         //printf("]\n");
      }else if(!strncmp("SON:",func_str,4)) {//Ultrasonic sensors
          int nson,count,nummax;
          static unsigned long int mask=0x00000000;
          unsigned  long int mask_aux,index;

          //printf("ULTRASONIC SENSORS\n");

          ros_gpmrc_state.son_active=true;
          ros_gpmrc_state.son_read=true;

          //printf("%s\n",func_str);
          nummax=sizeof(unsigned long int)*8;

          nson=atoi(dat_args[0]);//Number of ultrasonic sensors

          if (nson!=num_dat-2){
             printf("WRONG US MSG\n");
             return false;
          }
          ros_gpmrc_state.son_num=(int16_t)nson;

          mask_aux=(unsigned long int)atoi(dat_args[1]);//Sornars mask

          ros_gpmrc_state.son_mask=(uint16_t)mask_aux;

          //printf("Sonar mask: %X\t",mask_aux);
          index=0x1;
          for(i=0;i<nummax;i++){
              if ((index&mask_aux)&&(!(index&mask))){
                  mask|=index;
                  ros_gpmrc_state.son_num++;
              }
              index=index<<1;
          }

          index=0x1;
          count=0;
          ros_gpmrc_state.son_values.clear();
          for (i=0;i<nummax;i++){
            if (index&mask_aux){
              ros_gpmrc_state.son_values.push_back((int32_t)atoi(dat_args[2+count]));

              count++;
            }
            index=index<<1;
          }

         //printf("\n");
       }else if(!strncmp("BMP:",func_str,4)) {//Bumpers
          int nbmp;

          //printf("BUMPERS\n");

          ros_gpmrc_state.bmp_active=true;
          ros_gpmrc_state.bmp_read=true;

          nbmp=atoi(dat_args[0]);//Number of bumpers

/*
          if (nbmp!=num_dat-2){
             printf("WRONG BUMPERS MSG:%d\n",nbmp);
             return false;
          }
*/
          ros_gpmrc_state.bmp_num=(int16_t)nbmp;

          ros_gpmrc_state.bmp_state.clear();
          for (i=0;i<ros_gpmrc_state.bmp_num;i++){
            ros_gpmrc_state.bmp_state.push_back((bool)atoi(dat_args[1+i]));
          }
          //printf("]\n");
       }else if(!strncmp("REL:",func_str,4)) {//Relays state
           int nrel;

           ros_gpmrc_state.rel_active=true;
           ros_gpmrc_state.rel_read=true;

           nrel=atoi(dat_args[0]);//Number of relays
           if (nrel!=num_dat-1){
              printf("WRONG RELAYS MSG\n");
              return false;
           }

           ros_gpmrc_state.rel_num=(int16_t)nrel;
;
           ros_gpmrc_state.rel_state.clear();
           for (i=0;i< ros_gpmrc_state.rel_num;i++){
             ros_gpmrc_state.rel_state.push_back((bool)atoi(dat_args[1+i]));
           }
           //printf("]\n");
       }else if(!strncmp("ERR:",func_str,4)) {//Msg not recognized
            int err;

            err=atoi(dat_args[0]);//Error code
            printf("The controller didn't recognized the msg sent. ERROR:%d\n",err);
            //printf("]\n");
       }else if(!strncmp("ACK:",func_str,4)) {//Msg acknowledge
       }else printf("UNKNOWN ARGUMENT:%s\n",func_str);
       if (arg_str!=NULL) free(arg_str);
       if (func_str!=NULL) free(func_str);
  }else{
    printf("Unknown header byte:%c\n",Info[0]);
    return false;
  }

  return true;
}

int ExtractArguments(char *cad,char *array[],const char *separator){
  int loop,num;

  array[0]=strtok(cad,separator);
  if(array[0]==NULL){
    //printf("Din't find substrings.\n");
    return -1;
  }

  for(loop=1;loop<MAX_SUBCADENAS;loop++){//Extracts each substring
    array[loop]=strtok(NULL,separator);
    if(array[loop]==NULL)break;
  }
  num=loop;
/*
  g_print("Number of extracted substrings:%d\n",num);
  for(loop=0;loop<num;loop++){
    if(array[loop]==NULL) break;
    printf("Item #%d is %s.\n",loop,array[loop]);
  }
*/
  return num;
}
