#include "data_processor.h"


void DataProcessor::ReadConfigurationFile(char *filename){
    std::ifstream conf_file(filename);
    YAML::Parser parser(conf_file);
    YAML::Node doc;

    printf("Reading target map: %s...\n",filename);
    while(parser.GetNextDocument(doc)){
        if (const YAML::Node *pVal = doc.FindValue("RosFPS")){
          *pVal >> this->fps;
          printf("RosFPS:%f\n",this->fps);
        }
        if (const YAML::Node *pVal = doc.FindValue("RobotWidth")){
          *pVal >> this->robotWidth;
          printf("RosWidth:%f\n",this->robotWidth);
        }
        if (const YAML::Node *pVal = doc.FindValue("LaserLongPos")){
          *pVal >> this->laserLongPos;
          printf("LaserLongPos:%f\n",this->laserLongPos);
        }
        if (const YAML::Node *pVal = doc.FindValue("KPControl")){
          *pVal >> this->kpControl;
          printf("KPControl:%f\n",this->kpControl);
        }
        if (const YAML::Node *pVal = doc.FindValue("KDControl")){
          *pVal >> this->kdControl;
          printf("KDControl:%f\n",this->kdControl);
        }
        if (const YAML::Node *pVal = doc.FindValue("KIControl")){
          *pVal >> this->kiControl;
          printf("KIControl:%f\n",this->kiControl);
        }
        if (const YAML::Node *pVal = doc.FindValue("PlsAng")){
          *pVal >> this->plsAng;
          printf("PlsAng:%f\n",this->plsAng);
        }
        if (const YAML::Node *pVal = doc.FindValue("SteeringSemirangePls")){
          *pVal >> this->steerSemirangePls;
          printf("SteeringSemirangePls:%f\n",this->steerSemirangePls);
        }
        if (const YAML::Node *pVal = doc.FindValue("DoorMargin")){
          *pVal >> this->door_margin;
          printf("DoorMargin:%f\n",this->door_margin);
        }

    }
}
/*
void DataProcessor::PublishMark(path_follower_node_eusebio::Path pt_path)
{
    pointcloud_type mark_pcl_cloud;
    sensor_msgs::PointCloud2 msg;

    int numMarks;

    numMarks=1;
    //Point cloud
    mark_pcl_cloud.width = numMarks;
    mark_pcl_cloud.height =1;
    mark_pcl_cloud.is_dense = false;
    mark_pcl_cloud.points.resize(1);

    for(int j=0;j<(int)numMarks;j++){
       mark_pcl_cloud.points[j].x=pt_path.path[j].x;
       mark_pcl_cloud.points[j].y=pt_path.path[j].y;
       mark_pcl_cloud.points[j].z=0;
    }

    pcl::toROSMsg(mark_pcl_cloud, msg);
    string frame_id("/map");
    msg.header.stamp=this->timestamp;
    msg.header.frame_id=frame_id;

    //printf("Publishing markers cloud\n");
    this->pub_marks.publish(msg);
}
*/
void DataProcessor::PublishPath(void){
  //int size;
  nav_msgs::Path msg;
  geometry_msgs::PoseStamped path_pt;

  msg.header.frame_id="/map";

  msg.poses.clear();
  for (int i=0;i<(int)this->path.path.size();i++){
      path_pt.pose.position.x=this->path.path[i].x;
      path_pt.pose.position.y=this->path.path[i].y;
      path_pt.pose.position.z=0.0;
      path_pt.pose.orientation.x=0.0;
      path_pt.pose.orientation.y=0.0;
      path_pt.pose.orientation.z=0.0;
      path_pt.pose.orientation.w=0.0;
      msg.poses.push_back(path_pt);
  }
  this->pub_path.publish(msg);
}

void DataProcessor::PublishPathMarkers(void){
    visualization_msgs::Marker marker_path;

    //if (this->path.path.size()>2) RemovePathMarker(0);

    for (int i=0;i<(int)this->path.path.size();i++){
      marker_path.header.frame_id = "/map";
      marker_path.header.stamp = ros::Time::now();
      marker_path.ns = "path";
      marker_path.id = i;
      marker_path.type = visualization_msgs::Marker::SPHERE;

      marker_path.action = visualization_msgs::Marker::ADD;

      marker_path.pose.position.x = this->path.path[i].x;
      marker_path.pose.position.y = this->path.path[i].y;
      marker_path.pose.position.z = 0;

      marker_path.pose.orientation.x = 0.0;
      marker_path.pose.orientation.y = 0.0;
      marker_path.pose.orientation.z = 0.0;
      marker_path.pose.orientation.w = 1.0;

      marker_path.scale.x = 0.2;
      marker_path.scale.y = 0.2;
      marker_path.scale.z = 0.2;

      marker_path.color.r = 1.0f;
      marker_path.color.g = 0.0f;
      marker_path.color.b = 0.5f;
      marker_path.color.a = 1.0;
      marker_path.lifetime = ros::Duration();
      this->marker_pub.publish(marker_path);
   }
}

void DataProcessor::PublishPassedPathMarker( path_follower_node_eusebio::PathPoint point){
    visualization_msgs::Marker marker_path;
    static int count=0;

    marker_path.header.frame_id = "/map";
    marker_path.header.stamp = ros::Time::now();
    marker_path.ns = "passed_path";
    marker_path.id = count;
    marker_path.type = visualization_msgs::Marker::SPHERE;

    marker_path.action = visualization_msgs::Marker::ADD;

    marker_path.pose.position.x = point.x;
    marker_path.pose.position.y = point.y;
    marker_path.pose.position.z = 0;

    marker_path.pose.orientation.x = 0.0;
    marker_path.pose.orientation.y = 0.0;
    marker_path.pose.orientation.z = 0.0;
    marker_path.pose.orientation.w = 1.0;

    marker_path.scale.x = 0.2;
    marker_path.scale.y = 0.2;
    marker_path.scale.z = 0.2;

    marker_path.color.r = 0.5f;
    marker_path.color.g = 0.5f;
    marker_path.color.b = 0.5f;
    marker_path.color.a = 0.5f;
    marker_path.lifetime = ros::Duration();
    this->marker_pub.publish(marker_path);
    count++;
}

void DataProcessor::RemovePathMarker(int id){
    visualization_msgs::Marker marker_path;

    marker_path.header.frame_id = "/map";
    marker_path.header.stamp = ros::Time::now();
    marker_path.ns = "path";
    marker_path.id = id;
    marker_path.action = visualization_msgs::Marker::DELETE;

    this->marker_pub.publish(marker_path);
}

void DataProcessor::RemovePathMarkers(void){
    visualization_msgs::Marker marker_path;

    marker_path.header.frame_id = "/map";
    marker_path.header.stamp = ros::Time::now();
    marker_path.ns = "path";
    for (int i=0;i<(int)this->path.path.size();i++){
      marker_path.id = i;
      marker_path.action = visualization_msgs::Marker::DELETE;
      this->marker_pub.publish(marker_path);
    }
}

void DataProcessor::WritePathToFile(string filename){
  YAML::Emitter out;
  out << YAML::BeginSeq;

  out << YAML::BeginMap << YAML::Key << "path";
  out << YAML::Value << YAML::BeginSeq;

  for(int i=0;i< (int)this->path.path.size() ;i++){
     std::map <std::string, float>  node;

     out << YAML::Flow; //{.......}

     node["maneouver"] = this->path.path[i].maneouver;
     node["node_id"] = this->path.path[i].id;
     node["node_type"] = this->path.path[i].type;
     node["vel"] = this->path.path[i].lin_vel_max;
     node["x"] = this->path.path[i].x;
     node["y"] = this->path.path[i].y;
     node["theta"] = this->path.path[i].theta;
     out << node;
  }

  out << YAML::EndSeq << YAML::EndMap;
  out << YAML::EndSeq;

  std::ofstream fout(filename.c_str());
  fout << out.c_str();
}

void DataProcessor::LoadPathFromFile(string filename){
    std::ifstream fin(filename.c_str());

    YAML::Parser parser(fin);

    YAML::Node doc;

    while(parser.GetNextDocument(doc)) {
      const YAML::Node& nodeaction = doc[0];

      YAML::Iterator it=nodeaction.begin();

      std::string key;

      it.first() >> key;
      // std::cout << " Key["<< itask<< "] " << key <<  std::endl;

      const YAML::Node& nodepoint = nodeaction[key.c_str()];

      if(key.compare("path") == 0){
         for(unsigned i=0;i<nodepoint.size();i++) {
           path_follower_node_eusebio::PathPoint point;

           const YAML::Node& node = nodepoint[i];

           node["maneouver"] >>point.maneouver;
           node["node_id"] >>point.id;
           node["node_type"] >>point.type;
           node["vel"] >>point.lin_vel_max;
           node["x"] >>point.x;
           node["y"] >> point.y;
           node["theta"] >>point.theta;

           switch(point.maneouver){
             case FORWARD_MOVEMENT:
               printf("Loading path point:Type:PASS_POINT\t");
               break;
             case CHARGER_CONNECT:
               printf("Loading path point:Type:CHARGER_CONNECT_POINT\t");
               break;
             case CHARGER_DISCONNECT:
               printf("Loading path point:Type:CHARGER_DISCONNECT_POINT\t");
               break;
           }
           printf(" X:%f Y:%f\n",point.x,point.y);
           point.executed=0;//Point not executed yet
           this->path.path.push_back(point);
         }
      }
   }
}

void DataProcessor::PublishStatus(void){
    path_follower_node_eusebio::Status msg;

    //printf("publishing path follower status\n");
    if (!this->path.path.size()) msg.run_mode=NAV_STATUS_WAITTING_FOR_PATH;
    else if (!this->running_control) msg.run_mode=NAV_STATUS_NAVIGATION_PAUSED;
    else if (this->obstacle==2) msg.run_mode=NAV_STATUS_OBSTACLE;
    else msg.run_mode=NAV_STATUS_NAVIGATING;
    msg.current_point_id=this->point2.id;
    msg.dist=this->long_track_err;
    msg.alpha=this->steer_ref;
    this->pub_status.publish(msg);

}



void DataProcessor::missionAccomplished(void){
    std_msgs::String msg_command;

    printf("MISSION ACOMPLISHED\n");
    this->PublishPath();//Let it know it
    this->path.path.clear();
    this->maneouver_pending=false;
    this->mission_accomplished=true;
    this->path.idx=0;
    this->stopRobot();
/*
    printf("SENDING ABORT STAND_BY TO WII_IRODOM\n");
    msg_command.data="STAND_BY";
    this->pub_wii_command.publish(msg_command);     
    */
}

void DataProcessor::stopRobot(void){
    //printf("STOPPING ROBOT\n");
    this->joy_msg.axes[JOY_PC_AXIS_X1]=0.0;
    this->joy_msg.axes[JOY_PC_AXIS_Y1]=0.0;
    this->joy_msg.buttons[JOY_PC_BUTT_L2]=false;
    this->pub_joy_command.publish(joy_msg);
    this->running_control=false;
}



void DataProcessor::addCurrentPoseToPath(int type){
    path_follower_node_eusebio::PathPoint point;

    printf("ADDING CURRENT POSE AS TYPE:%d\n",type);

    point.id=(int)(this->path.path.size());
    point.type=NO_ACKNOWLEDGE;
    point.lin_vel_max=DEFAULT_LIN_VEL_MAX;
    point.x=this->current_pose.xs;
    point.y=this->current_pose.ys;
    point.theta=this->current_pose.theta;
    point.maneouver=type;
    point.executed=0;
    this->path.path.push_back(point);
    this->PublishPath();
}


void DataProcessor::getGapDirection(void){
   //printf("Num gaps:%d\n",(unsigned int)gap_list.size());
   this->gap_center=M_PI;

   for (int j=0;j<(int)this->gap_list.size();j++){
     printf("Gap %d: Left:(X:%f Ang:%f Dist:%f) Right:(X:%f Ang:%f Dist:%f) Center_angle:%f Gap:%f\n",j,gap_list[j].x_left,gap_list[j].left_ang*180.0/M_PI, gap_list[j].dist_left, gap_list[j].x_right,gap_list[j].right_ang*180.0/M_PI, gap_list[j].dist_right ,gap_list[j].center_ang*180.0/M_PI, gap_list[j].width);
     if (fabs(this->gap_list[j].center_ang)<fabs(this->gap_center)) this->gap_center=this->gap_list[j].center_ang;
   }
   printf("Most centered gap: %f\n",this->gap_center*180.0/M_PI);
   printf(" Steering ang:%f\n",this->steer_ang*180.0/M_PI);
   //To avoid excesive steering control due to noise in the gap detection
   //if (this->gap_center>MAX_STEER_DOOR_ANGLE) this->gap_center=0;
   //else if (this->gap_center<-MAX_STEER_DOOR_ANGLE) this->gap_center=0;
}

void DataProcessor::detectObstaculization(sensor_msgs::LaserScan msg_laser){
   double angle_min,angle_max,angle_increment;
   unsigned int idx_min,idx_max,num_readings;

   //------------------ OBSTACLE DETECTION ------------
   angle_min=(msg_laser.angle_max+msg_laser.angle_min)/2.0-LASER_BUMPER_ANGLE_RANGE/2.0+this->steer_ang;
   if (angle_min<msg_laser.angle_min) angle_min=msg_laser.angle_min;
   else if (angle_min>msg_laser.angle_max) angle_min=msg_laser.angle_max;

   angle_max=(msg_laser.angle_max+msg_laser.angle_min)/2.0+LASER_BUMPER_ANGLE_RANGE/2.0+this->steer_ang;
   if (angle_max<msg_laser.angle_min) angle_max=msg_laser.angle_min;
   if (angle_max>msg_laser.angle_max) angle_max=msg_laser.angle_max;

   //printf(" Steering ang:%f Angle min:%f Angle max:%f\n",msg_robot_status.steering_ang*180.0/M_PI,angle_min*180.0/M_PI,angle_max*180.0/M_PI);

   angle_increment=msg_laser.angle_increment;
   num_readings=(unsigned int)((angle_max-angle_min)/(angle_increment));
   idx_min=(angle_min-msg_laser.angle_min)/msg_laser.angle_increment;
   idx_max=idx_min+num_readings;
   this->obstacle=false;
   for (unsigned int i=0;i<num_readings;i++) {
     if (msg_laser.ranges[i+idx_min] && (msg_laser.ranges[i+idx_min]<MIN_LASER_BUMPER_DISTANCE+this->vel)) {
       this->obstacle=true;
       //printf("i:%d Dist:%f, ",i,msg_laser.ranges[i+idx_min]);
     }
   }

   //if (this->obstacle) printf ("OBSTACLE...\n");
}


void DataProcessor::detectObstacles(sensor_msgs::LaserScan msg_laser){
  double angle_left_edge,x_left_edge, z_left_edge, angle_right_edge,x_right_edge, z_right_edge, dist_edge, dist_right_edge,dist_left_edge;
  int i,num_readings;
  Obstacle obstacle_data;

  num_readings=(unsigned int)((msg_laser.angle_max-msg_laser.angle_min)/(msg_laser.angle_increment));
  //printf("Num readings:%d, Ang_min:%f, Ang_max:%f, Ang_incr:%f\n",num_readings,msg_laser.angle_min*180.0/M_PI,msg_laser.angle_max*180.0/M_PI,msg_laser.angle_increment*180.0/M_PI);
  //printf("---------------------------------- OBSTACLES ------------------------------------------\n");
  this->obstacle_list.clear();

  i=0;
  while(i<num_readings){  
   //Search door left edge start
    do{
       dist_edge=msg_laser.ranges[i];
       if (dist_edge<MIN_DIST_LASER) dist_edge=MAX_DIST_LASER;
       i++;
    }while ((i<num_readings)&&(dist_edge>MAX_OBSTACLE_EDGE_DIST));

    if (i>=num_readings) return;

    i--;

    x_right_edge=MAX_OBSTACLE_EDGE_DIST;
    x_left_edge=-MAX_OBSTACLE_EDGE_DIST;

    while((i<num_readings)&&(dist_edge<=MAX_OBSTACLE_EDGE_DIST)){
        double count;
        double x_obst;
        double angle;

        dist_edge=msg_laser.ranges[i];
        if (dist_edge<MIN_DIST_LASER) dist_edge=MAX_DIST_LASER;

        if (dist_edge<=MAX_OBSTACLE_EDGE_DIST){
          count=(double)i;
          angle=(double)(count*msg_laser.angle_increment)+(double)msg_laser.angle_min;
          x_obst=dist_edge*sin(angle);

          if (x_obst < x_right_edge) {
            //printf("%d Dist:%f X:%f\n",i,dist_edge,x_obst);
            x_right_edge=x_obst;
            z_right_edge=dist_edge*cos(angle)+this->laserLongPos;                  
            dist_right_edge=dist_edge;
            angle_right_edge=angle;
          }

          if (x_obst > x_left_edge) {
            x_left_edge=x_obst;
            z_left_edge=dist_edge*cos(angle)+this->laserLongPos;                  
            dist_left_edge=dist_edge;
            angle_left_edge=angle;
          }
        }
        i++;
    }
    i--;

    obstacle_data.left_ang=angle_left_edge;
    obstacle_data.right_ang=angle_right_edge;
    obstacle_data.dist_left=dist_left_edge;
    obstacle_data.dist_right=dist_right_edge;
    obstacle_data.x_left=x_left_edge;
    obstacle_data.x_right=x_right_edge;
    obstacle_data.z_left=z_left_edge;
    obstacle_data.z_right=z_right_edge;
    this->obstacle_list.push_back(obstacle_data);
    //printf("Obstacle  Left:(X:%f Ang:%f Dist:%f) Right:(X:%f Ang:%f Dist:%f)\n",obstacle_data.x_left,obstacle_data.left_ang*180.0/M_PI, obstacle_data.dist_left, obstacle_data.x_right,obstacle_data.right_ang*180.0/M_PI, obstacle_data.dist_right);
  }
}

void DataProcessor::detectGaps(void){
    Gap gap_data;

    //printf("---------------------------------- GAPS ------------------------------------------\n");

    this->gap_list.clear();
    this->gap_exist=false;

    if (!obstacle_list.size()) {
      gap_data.left_ang=103.0*M_PI/180.0;
      gap_data.dist_left=MAX_OBSTACLE_EDGE_DIST;
      gap_data.x_left=gap_data.dist_left*sin(gap_data.left_ang);
      gap_data.z_left=gap_data.dist_left*cos(gap_data.left_ang);
      gap_data.right_ang=-103.0*M_PI/180.0;
      gap_data.dist_right=MAX_OBSTACLE_EDGE_DIST;
      gap_data.x_right=gap_data.dist_right*sin(gap_data.right_ang);
      gap_data.z_right=gap_data.dist_right*cos(gap_data.right_ang);
      gap_data.width=gap_data.x_left-gap_data.x_right;
      //printf("Gap: Left:(X:%f Ang:%f Dist:%f) Right:(X:%f Ang:%f Dist:%f) Center_angle:%f Gap:%f\n",gap_data.x_left,gap_data.left_ang*180.0/M_PI, gap_data.dist_left, gap_data.x_right,gap_data.right_ang*180.0/M_PI, gap_data.dist_right,gap_data.center_ang*180.0/M_PI,gap_data.width);

      gap_list.push_back(gap_data);
      this->gap_exist=true;
      return;
    }

    for (int i=0;i<=(int)obstacle_list.size();i++){
       if (this->getGap(i,i-1,&gap_data)) {
         gap_list.push_back(gap_data);
         this->gap_exist=true;
       }
    }
}

int DataProcessor::getGap(int idx, int idx_prev, Gap *gap_data){
  int nobst,ret;
  double min_gap;
  double alpha_max,alpha_min;

  ret=0;
  nobst=(int)obstacle_list.size();
  min_gap=this->robotWidth+2.0*this->door_margin;

  if (idx_prev<0){//Analize gap before first obstacle
    gap_data->left_ang=obstacle_list[idx].right_ang;
    gap_data->dist_left=obstacle_list[idx].dist_right;
    gap_data->x_left=obstacle_list[idx].x_right;
    gap_data->z_left=obstacle_list[idx].z_right;
    gap_data->right_ang=-103.0*M_PI/180.0;
    gap_data->dist_right=MAX_OBSTACLE_EDGE_DIST;
    gap_data->x_right=gap_data->dist_right*sin(gap_data->right_ang);
    gap_data->z_right=gap_data->dist_right*cos(gap_data->right_ang);
  }else if (idx<nobst){//Gaps between obstacles
    gap_data->left_ang=obstacle_list[idx].right_ang;
    gap_data->dist_left=obstacle_list[idx].dist_right;
    gap_data->x_left=obstacle_list[idx].x_right;
    gap_data->z_left=obstacle_list[idx].z_right;
    gap_data->right_ang=obstacle_list[idx-1].left_ang;
    gap_data->dist_right=obstacle_list[idx-1].dist_left;
    gap_data->x_right=obstacle_list[idx-1].x_left;
    gap_data->z_right=obstacle_list[idx-1].z_left;
  }else{//Gap after the last obstacle
    gap_data->left_ang=103.0*M_PI/180.0;
    gap_data->dist_left=MAX_OBSTACLE_EDGE_DIST;
    gap_data->x_left=gap_data->dist_left*sin(gap_data->left_ang);
    gap_data->z_left=gap_data->dist_left*cos(gap_data->left_ang);
    gap_data->right_ang=obstacle_list[idx_prev].left_ang;
    gap_data->dist_right=obstacle_list[idx_prev].dist_left;
    gap_data->x_right=obstacle_list[idx_prev].x_left;
    gap_data->z_right=obstacle_list[idx_prev].z_right;
  }
  gap_data->width=gap_data->x_left-gap_data->x_right;

  if (gap_data->x_left < (min_gap/2.0)){//Obstacle in the left
    alpha_max=atan2(gap_data->x_left,gap_data->z_left);
    if ((gap_data->x_left-gap_data->x_right) >= min_gap) {//The robot fits in the gap
      alpha_min=atan2(gap_data->x_left-min_gap*1.1,gap_data->z_left);
      ret=1;
    }else{//The robot doesn't fit in the gap
      alpha_min=atan2(gap_data->x_right,gap_data->z_right);
      ret=0;
    }
    gap_data->center_ang=(alpha_max-alpha_min)/2.0+alpha_min;
  }else if (gap_data->x_right > -(min_gap/2.0)) {//Obstacle in the right
    alpha_min=atan2(gap_data->x_right,gap_data->z_right);
    if ((gap_data->x_left-gap_data->x_right) >= min_gap) {//The robot fits in the gap
      alpha_max=atan2(gap_data->x_right+min_gap*1.1,gap_data->z_right);
      ret=1;
    }else{
      alpha_min=atan2(gap_data->x_right,gap_data->z_right);
      ret=0;
    }
    gap_data->center_ang=(alpha_max-alpha_min)/2.0+alpha_min;
  }else{
    gap_data->center_ang=0.0;
    ret=1;
  }
    //printf("Gap: Left:(X:%f Ang:%f Dist:%f) Right:(X:%f Ang:%f Dist:%f) Center_angle:%f Gap:%f\n",gap_data->x_left,gap_data->left_ang*180.0/M_PI, gap_data->dist_left, gap_data->x_right,gap_data->right_ang*180.0/M_PI, gap_data->dist_right,gap_data->center_ang*180.0/M_PI,gap_data->width);

  return ret;
}


void DataProcessor:: maneouverExecuted(unsigned int maneouver_type){
  std_msgs::String msg_command;
/*
  switch(maneouver_type){
     case CHARGER_DISCONNECT:
       printf("SENDING ABORT STAND_BY TO WII_IRODOM\n");
       msg_command.data="STAND_BY";
       this->pub_wii_command.publish(msg_command);
       this->maneouver_pending=false;//We finished the disconnect maneouver
       this->nextPathPoint();//We take the next point if any
       break;
     case THROUGHT_DOOR_MOVEMENT:
       this->dist_door_travelled=0.0;
     case CHARGER_CONNECT:
     case FORWARD_MOVEMENT:
       this->maneouver_pending=false;//We finished the connect maneouver
       this->nextPathPoint();//We take the next point if any
       break;
       break;
  }
  */
}

void DataProcessor:: getTrackError(void){
  double dx,dy;
  double drx,dry,length_segment;
  double alphar,alpha12,incr_alpha;

  dx=this->point2.x-this->point1.x;
  dy=this->point2.y-this->point1.y;
  length_segment=sqrt(pow(dx,2.0)+pow(dy,2.0));
  alpha12=atan2(dy,dx);

  drx=this->current_pose.xs-this->point1.x;
  dry=this->current_pose.ys-this->point1.y;

  alphar=this->current_pose.theta;

  incr_alpha=alphar-alpha12;

  this->ori_track_err=this->moduleRange(incr_alpha,-M_PI, M_PI)/M_PI;

  if (length_segment>=MIN_LENGTH_SEGMENT){
    //this->long_track_err is the robot estimate projected onto the path segment
    this->long_track_err=(drx*dx+dry*dy)/pow(length_segment,2.0);

    //this->cross_track_err is the estimate projected onto the normal of the path segment
    this->cross_track_err=(dry*dx-drx*dy)/length_segment/MAX_CROSS_TRACK_ERR;

    if (this->cross_track_err>1.0) this->cross_track_err=1.0;
    else if (this->cross_track_err<-1.0) this->cross_track_err=-1.0;

  }else{
    this->cross_track_err=0.0;
    this->long_track_err=1.0;
    this->ori_track_err=0.0;
  }
}

void DataProcessor::nextPathPoint(void){
  this->path.idx++;
  this->path.path[this->path.idx-1].executed=1;//Path point executed
  if (this->path.idx>=(int)this->path.path.size()){
    this->missionAccomplished();
  }else{//Take next segment and update maneouver_pending flag to false and the speed limits
    if ((this->path.path[this->path.idx].maneouver==CHARGER_DISCONNECT)||(this->path.path[this->path.idx].maneouver==CHARGER_CONNECT)){
      this->point1=this->path.path[this->path.idx];
      this->point2=this->point1;
    }else{
      this->point2=this->path.path[this->path.idx];
      this->point1=this->path.path[this->path.idx-1];
    }
    this->point_reached=false;
    this->maneouver_pending=false;
    this->max_lin_vel=this->point2.lin_vel_max;
    this->min_lin_vel=DEFAULT_LIN_VEL_MIN;
    this->running_control=true;
  }
}

void DataProcessor::calculateReferences(double *vel_ref, double *steer_ref){
  if (this->maneouver_pending==true){
    switch (this->point2.maneouver){
      case THROUGHT_DOOR_MOVEMENT:
        (*steer_ref)=this->gap_center;
        if (this->gap_exist){
           (*vel_ref)=this->min_lin_vel; 
           //-------- PROVISIONAL --------
           //(*vel_ref)=0.0;
        }
        else {
           printf("No gap to continue. Stopping..\n");
           (*vel_ref)=0.0;
        }
        break;
      case FORWARD_MOVEMENT:
        break;
     }
  }else{//Point2 not reached yet
     double diff_cte;
     static double int_cte=0.0;
     double track_err;

     track_err=(WEIGHT_CROSS_ERR*this->cross_track_err+WEIGHT_ORI_ERR*this->ori_track_err);

     diff_cte = track_err-this->track_err_prev;
     int_cte+=track_err;

     this->track_err_prev=track_err;

     (*steer_ref) = - this->kpControl * track_err - this->kdControl * diff_cte - this->kiControl * int_cte; 

     if ((*steer_ref)>this->steeringSemiRangeRad) (*steer_ref)=this->steeringSemiRangeRad;
     else if ((*steer_ref)<-this->steeringSemiRangeRad) (*steer_ref)=-this->steeringSemiRangeRad;

     this->steer_ang_err=(*steer_ref)-this->steer_ang;
     if (this->steer_ang_err > this->steeringSemiRangeRad){
       this->steer_ang_err=this->steeringSemiRangeRad;
     }else if (this->steer_ang_err < -this->steeringSemiRangeRad){
       this->steer_ang_err=-this->steeringSemiRangeRad;
     }

     (*vel_ref)=this->max_lin_vel*fabs(this->steeringSemiRangeRad-this->steer_ang_err)/this->steeringSemiRangeRad;
  }
}

double DataProcessor::moduleRange(double value, double lim_min, double lim_max){
  if (value>lim_max) value=lim_min+(value-lim_max);
  else if (value<=lim_min) value=lim_max-(lim_min-value);
  return value;
}

void DataProcessor::sendReferences(double vel,double steer){
    float x_val,y_val;

    y_val=vel;
    x_val=steer/this->steeringSemiRangeRad;

    //printf("CONTROL REF: Speed:%f Steering angle:%f\n",vel,steer*180.0/M_PI);
    if (y_val>1.0) y_val=1.0;
    else if (y_val<-1.0) y_val=-1.0;
    if (x_val>1.0) x_val=1.0;
    else if (x_val<-1.0) x_val=-1.0; 

    //printf("JOY REFERENCES: X:%f Y:%f\n",x_val,y_val);

    this->joy_msg.axes[JOY_PC_AXIS_X1]=x_val;
    if (this->obstacle){
      this->joy_msg.axes[JOY_PC_AXIS_Y1]=0.0;
      this->joy_msg.buttons[JOY_PC_BUTT_L2]=false;
    }else{
      this->joy_msg.buttons[JOY_PC_BUTT_L2]=true;
      this->joy_msg.axes[JOY_PC_AXIS_Y1]=y_val;
    }
    this->pub_joy_command.publish(joy_msg);
}

