#ifndef MY_IMAGE_PROCESOR_H
#define MY_IMAGE_PROCESSOR_H

//#define USE_KALMAN_FILTER_STATUS
#define USE_ROBOT_STATUS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <sstream>
#include <vector>

//OpenCV
#include </opt/ros/groovy/include/opencv/cv.h>

//ROS
#include "ros/ros.h"
#include <geometry_msgs/PoseWithCovarianceStamped.h>

#include <std_msgs/String.h>
#include <std_msgs/Bool.h>
#include <std_msgs/Int8.h>
#include <std_msgs/Float64.h>

#include <sensor_msgs/LaserScan.h>

/*
//To publish point clouds of point
#include <sensor_msgs/PointCloud2.h>
#include <pcl/io/io.h>
#include <pcl/io/pcd_io.h>
#include <pcl/registration/transforms.h>
*/
#include "nav_msgs/Path.h"

//YAML reading
#include <fstream>
#include "yaml-cpp/yaml.h"
#include <vector>

//Service topics
#include <std_srvs/Empty.h>

#include "path_follower_node_eusebio/Goal.h"
#include "path_follower_node_eusebio/Status.h"
#include "path_follower_node_eusebio/Path.h"
#include "path_follower_node_eusebio/PathPoint.h"

//#include "wii_irodom/Status.h"

//Synchronization
#include "message_filters/subscriber.h"
#include "message_filters/time_synchronizer.h"
#include "message_filters/sync_policies/approximate_time.h"
#include "message_filters/sync_policies/exact_time.h"

#include <visualization_msgs/Marker.h>

#include <sensor_msgs/Joy.h>

#ifdef USE_ROBOT_STATUS
  #include "eusebio/RobotStatus.h"
#elif defined USE_KALMAN_FILTER_STATUS
  #include "kalman_filter_node/Status.h"
#endif

//#define GUIDE_POINT_IDX 1
#define MIN_DIST_GOAL 0.1 //Distance limit to the final goal to make lineal speed zero
#define MIN_DIST_LOCAL1 0.3 //Distance limit to the guide point to reduce speed linearly
#define MIN_DIST_LOCAL2 0.2 //Distance limit to the guide point to make lineal speed zero
#define MIN_DIST_LOCAL3 0.5 //Distance limit to the guide point to make lineal speed zero

#define GOAL_ORI_THRESHOLD 0.05 //Final goal angular error in radians

#define MIN_NUM_JOY_AXES 6
#define MIN_NUM_JOY_BUTTONS 13

//Buttons
#define JOY_PC_BUTT_SQUARE     0
#define JOY_PC_BUTT_CROSS      1
#define JOY_PC_BUTT_CIRCLE     2
#define JOY_PC_BUTT_TRIANGLE   3
#define JOY_PC_BUTT_L2         6
#define JOY_PC_BUTT_SEL       8

//Axes
#define JOY_PC_AXIS_X1               0
#define JOY_PC_AXIS_Y1               1
#define JOY_PC_BUTT_LEFT_RIGHT       4
#define JOY_PC_BUTT_UP_DOWN          5

//Maneouvers
#define FORWARD_MOVEMENT  	30
#define BACKWARD_MOVEMENT 	31
#define CHARGER_CONNECT   	33
#define CHARGER_DISCONNECT   	34
#define THROUGHT_DOOR_MOVEMENT 	35

//Node types
#define WITH_ACKNOWLEDGE 0
#define NO_ACKNOWLEDGE 1

//Default node values
#define DEFAULT_LIN_VEL_MAX 1.0
#define DEFAULT_LIN_VEL_MIN 0.8


//Run modes
#define NAV_STATUS_POS_NOT_CALIBRATED   0
#define NAV_STATUS_WAITTING_FOR_PATH    1
#define NAV_STATUS_NAVIGATION_PAUSED    2
#define NAV_STATUS_OBSTACLE             3
#define NAV_STATUS_NAVIGATING           4

#define VEL_FILTER_COEF 0.2

//WII IRODOM Run modes
#define WII_IRODOM_RUN_MODE_WAITING            0
#define WII_IRODOM_RUN_MODE_SEARCHING          1
#define WII_IRODOM_RUN_MODE_GOTO_TARGET_PH1    2
#define WII_IRODOM_RUN_MODE_GOTO_TARGET_PH2    3
#define WII_IRODOM_RUN_MODE_GOTO_TARGET_PH3    4
#define WII_IRODOM_RUN_MODE_GOFROM_TARGET_PH2  5
#define WII_IRODOM_RUN_MODE_GOFROM_TARGET_PH3  6
#define WII_IRODOM_RUN_MODE_CONNECTED          7
#define WII_IRODOM_RUN_MODE_DISCONNECTED       8

#define MAX_LASER_BUMPER_DISTANCE 0.8
#define MIN_LASER_BUMPER_DISTANCE 0.3
#define LASER_BUMPER_ANGLE_RANGE (M_PI/4.0)

#define DOOR_EDGE_NONE 0
#define DOOR_EDGE_LEFT 1
#define DOOR_EDGE_RIGHT 2

#define MAX_OBSTACLE_EDGE_DIST 1.0

#define DEFAULT_ROBOT_WIDTH 0.65
#define DEFAULT_LASER_LONG_POS 0.18

#define DEFAULT_DOOR_MARGIN 0.03
#define MAX_DOOR_TRAVEL_DIST 2.0
#define MAX_STEER_DOOR_ANGLE 0.5

#define MIN_DIST_LASER 0.1
#define MAX_DIST_LASER 10.0

//Controller
#define DEFAULT_KP_CONTROL 2.0
#define DEFAULT_KD_CONTROL 0.2
#define DEFAULT_KI_CONTROL 0.000

#define MIN_LENGTH_SEGMENT 0.1

#define WEIGHT_CROSS_ERR 0.25
#define WEIGHT_ORI_ERR 0.5

#define MAX_CROSS_TRACK_ERR 0.2

typedef struct _Pose{
    //Steering coordinates
    double xs;
    double ys;

    //Orientation
    double theta;
}Pose;

typedef struct _Gap{

  double left_ang;
  double dist_left;
  double x_left;
  double z_left;
  double right_ang;
  double dist_right;
  double x_right;
  double z_right;
  double center_ang;
  double width;
}Gap;

typedef struct _Obstacle{
  double left_ang;
  double dist_left;
  double x_left;
  double z_left;
  double right_ang;
  double dist_right;
  double x_right;
  double z_right;
}Obstacle;

/*
typedef pcl::PointXYZRGB point_type;
//typedef pcl::PointXYZ point_type;
typedef pcl::PointCloud<point_type> pointcloud_type;
*/
using namespace std;
using namespace message_filters;

//Class for processing of the incoming images
class DataProcessor
{

public:
  ros::Time timestamp;//Last timestamp in ROS format
  double tstamp;//Timestamp

  double pitch,roll,yaw;

  Pose current_pose;

  float fps;

  path_follower_node_eusebio::Path path;

  //Publishers
  ros::Publisher pub_marks,pub_joy_command,marker_pub, pub_path;
  ros::Publisher pub_status;
  //ros::Publisher pub_wii_command;

  path_follower_node_eusebio::PathPoint point1,point2; //Points of the path segment to follow

  //joy::Joy joy_msg; //old version
  sensor_msgs::Joy joy_msg; //new ROS joy topic version
  ros::NodeHandle node_handle;

  double vel_ref,steer_ref;//References
  double cross_track_err,long_track_err,ori_track_err;//Error to the path segment
  double track_err_prev;
  double vel,steer_ang;//Current cinematic values
  double steer_ang_err;//steer ang error
  double max_lin_vel,min_lin_vel;//Vel linear limits
  double ang_err;

  //Useful flags
  bool running_control;//Whether the control is currently running or paused
  bool point_reached;//Whether the next point has been reached or we are still going to it
  bool maneouver_pending;//Whether there is a maneouver currently in execution or not
  bool mission_accomplished;//Whether the whole path mission has been executed (accomplished) or not
  bool obstacle;//Whether there is an obstacle in the path or not
  bool position_calibrated;//Whether the position of the robot has been initialized (calibrated) or not

  //Some robot geometry parameters
  double robotWidth;
  double laserLongPos;

  //Obstacle avoidance (for door crossing)
  double gap_center;
  double dist_door_travelled;
  bool gap_exist;
  vector<Gap> gap_list;
  vector<Obstacle> obstacle_list;

  double kpControl,kdControl,kiControl;
 
  double plsAng,steerSemirangePls,steeringSemiRangeRad;
  double door_margin;

  DataProcessor(void)//Constructor
  {
      fps=30;
      path.idx=0;
      running_control=false;
      point_reached=true;
      maneouver_pending=false;
      mission_accomplished=true;
      obstacle=0;
      gap_exist=false;
      obstacle=false;
      position_calibrated=false;

      max_lin_vel=DEFAULT_LIN_VEL_MAX;
      min_lin_vel=DEFAULT_LIN_VEL_MIN;
      for (int i=0;i<MIN_NUM_JOY_BUTTONS;i++){
          joy_msg.buttons.push_back(0);
      }
      for (int i=0;i<MIN_NUM_JOY_AXES;i++){
          joy_msg.axes.push_back(0);
      }
      robotWidth=DEFAULT_ROBOT_WIDTH;
      laserLongPos=DEFAULT_LASER_LONG_POS;
      dist_door_travelled=0.0;
      track_err_prev=0.0;

      kpControl=DEFAULT_KP_CONTROL;
      kdControl=DEFAULT_KD_CONTROL;
      kiControl=DEFAULT_KI_CONTROL;
      door_margin=DEFAULT_DOOR_MARGIN;
  }

  ~DataProcessor(void)//Destructor
  {
  }  

  void ReadConfigurationFile(char *filename);
 // void PublishMark(path_follower_node_eusebio::Path pt_path);
  void PublishPath(void);
  void PublishRobotMarker(void);
  void PublishPathMarkers(void);
  void RecalculatePathToCurrentGoal(void);
  void RemovePathMarker(int id);
  void PublishPassedPathMarker( path_follower_node_eusebio::PathPoint point);
  void WritePathToFile(string filename);
  void LoadPathFromFile(string filename);
  void RemovePathMarkers(void);
  void PublishStatus(void);
  void missionAccomplished(void);
  void calculateReferences(double *vel_ref, double *steer_ref);
  void nextPathPoint(void);
  void stopRobot(void);
  void sendReferences(double vel,double steer);
  void addCurrentPoseToPath(int type);
  //void detectGaps(sensor_msgs::LaserScan msg_laser);
  void detectGaps(void);
  void getGapDirection(void);
  void detectObstaculization(sensor_msgs::LaserScan msg_laser);
  void detectObstacles(sensor_msgs::LaserScan msg_laser);
  int getGap(int idx, int idx_prev, Gap *gap_data);
  void maneouverExecuted(unsigned int maneouver_type);
  void getTrackError(void);
  double moduleRange(double value, double lim_min, double lim_max);
};

#endif // MY_IMAGE_PROCESSOR_H


