/************************************************************************
 * by Salvador Dominguez (from CARTIF Foundation)
 *
 * Copyright (C) 2011
 *
 * Salvador Dominguez Quijada
 * saldom@cartif.es
 *
 * Path_follower_agv is licenced under the GPL License
 *
 * You are free:
 *   - to Share - to copy, distribute and transmit the work
 *   - to Remix - to adapt the work
 *
 * Under the following conditions:
 *
 *   - Attribution. You may mention the Nifti project and the authors
 *
 *   - Noncommercial. You may not use this work for commercial purposes.
 *
 *   - Share Alike. If you alter, transform, or build upon this work,
 *     you may distribute the resulting work only under the same or
 *     similar license to this one.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 ************************************************************************/

/**
  @file path_follower_agv.cpp
  @brief ROS node to generate /joy references to follow a defined path composed by maneouvers
  @author Salvador Dominguez
  @date 9/8/2013
  */

#include "data_processor.h"

//Namespaces
using namespace std;
using namespace cv;
using namespace sensor_msgs;

//A pointer to the main data processor object
DataProcessor *dp=0;

/**
  @brief Callback to indicate WiiIR odom maneouvers finishing, so the path follower can continue with the next point int the path
  @param msg wii_irodom::Status ROS topic with the status of the wii_irodom node. Defined in the wii_irodom node
  */
/*
void WiiIROdomCallback(wii_irodom::Status msg){
  //printf("WII_IRODOM STATUS:%d PATH_SIZE:%d GOAL_REACHED:%d\n",msg.run_mode,dp->path.path.size());

  if (dp->maneouver_pending==true){
    switch(dp->point2.maneouver){
       case CHARGER_DISCONNECT:
          if (msg.run_mode==WII_IRODOM_RUN_MODE_DISCONNECTED){
            printf("DISCONNECT MANEOUVER FINISHED\n");
            dp->maneouverExecuted(CHARGER_DISCONNECT);
          }
          break;
       case CHARGER_CONNECT:
          if (msg.run_mode==WII_IRODOM_RUN_MODE_CONNECTED){
            printf("CONNECT MANEOUVER FINISHED\n");
            dp->maneouverExecuted(CHARGER_CONNECT);
          }
          break;
    }
  }
}
*/
/**
  @brief Callback to receive commands from other modules
  @param msg std_msgs::Sring ROS topic with the coded command to execute
  */
void CommandCallback(std_msgs::String msg){
  //ROS_INFO("PATH_FOLLOWER_COMMAND comming:%s\n",msg.data.c_str());
  if (!strcmp(msg.data.c_str(),"ABORT")){//Abort current path execution
    //printf("ABORTING PATH FOLLOWING\n");
    dp->missionAccomplished();
  }else if (!strcmp(msg.data.c_str(),"FOLLOW_PATH")){//Follow a path or pause current execution
    if (dp->running_control){
      printf("PAUSING MOVEMENT\n");
      dp->running_control=false;
      dp->stopRobot();
    }else{
      printf("FOLLOWING PATH\n");
      printf("Path size:%d\n",(int)dp->path.path.size());
      if (dp->path.path.size()) {
        if (dp->path.idx==0){
          dp->point2=dp->path.path[0];
          dp->point1=dp->point2;
          dp->point1.x=dp->current_pose.xs;//X coordinate of the node in meters
          dp->point1.y=dp->current_pose.ys;//Y coordinate of the node in meters
          dp->point1.theta=dp->current_pose.theta;//Orientation of the node in radians
        }else{
          if ((dp->path.path[dp->path.idx].maneouver==CHARGER_DISCONNECT)||(dp->path.path[dp->path.idx].maneouver==CHARGER_CONNECT)){
            dp->point1=dp->path.path[dp->path.idx];
            dp->point2=dp->point1;
          }else{
            dp->point2=dp->path.path[dp->path.idx];
            dp->point1=dp->path.path[dp->path.idx-1];
          }
        }
        dp->mission_accomplished=false;
        dp->running_control=true;//Execute current goal
        dp->point_reached=false;
        dp->maneouver_pending=false;
        dp->max_lin_vel=dp->point2.lin_vel_max;
        dp->min_lin_vel=DEFAULT_LIN_VEL_MIN;
      }else{
        dp->running_control=false;
        dp->mission_accomplished=true;
        dp->maneouver_pending=false;
        dp->point_reached=true;
      }
    }
  }else if (!strncmp(msg.data.c_str(),"SAVE_PATH",9)){//Save current path into an yaml file
    if (dp->path.path.size()){//If path not empty
      string filename_str;

      filename_str=msg.data.substr(10,strlen(msg.data.c_str())-1);

      printf("SAVING PATH to file:%s\n",filename_str.c_str());
      dp->WritePathToFile(filename_str);
    }//else printf("NO PATH TO SAVE\n");
  }else if (!strncmp(msg.data.c_str(),"LOAD_PATH",9)){//Load a path from a yaml file
    string filename_str;

    filename_str=msg.data.substr(10,strlen(msg.data.c_str())-1);
    //printf("LOADING PATH from file:%s\n",filename_str.c_str());
    //First remove current path if any
    dp->path.path.clear();
    //Load the path
    dp->LoadPathFromFile(filename_str);
    dp->PublishPath();
    dp->path.idx=0;
  }else if (!strcmp(msg.data.c_str(),"ADD_PASS_POINT")){//Adding current position as a pass path point
    dp->addCurrentPoseToPath(FORWARD_MOVEMENT);
  }else if (!strcmp(msg.data.c_str(),"ADD_CHARGER_CONNECT_POINT")){//Adding current position as a connect-to-charger point
    dp->addCurrentPoseToPath(CHARGER_CONNECT);
  }else if (!strcmp(msg.data.c_str(),"ADD_CHARGER_DISCONNECT_POINT")){//Adding a disconnect-from-charger point
    dp->addCurrentPoseToPath(CHARGER_DISCONNECT);
  }else if (!strcmp(msg.data.c_str(),"ADD_THROUGHT_DOOR_POINT")){//Add current position as a pass throught door point
    dp->addCurrentPoseToPath(THROUGHT_DOOR_MOVEMENT);
  }
}

void LaserSyncCallback(eusebio::RobotStatus msg_robot_status, sensor_msgs::LaserScan msg_laser){
//void LaserSyncCallback(kalman_filter_node::Status msg_robot_status, sensor_msgs::LaserScan msg_laser){
/*   double deltaT;
   double t_now;
   static double t_ant=0.0;
   
   //printf("Sync robot_status-laser\n");

   //--------------- Odom status update ------------
   dp->position_calibrated=msg_robot_status.position_calibrated;
   dp->current_pose.xs=msg_robot_status.xs;
   dp->current_pose.ys=msg_robot_status.ys;
   //dp->current_pose.xc=msg_robot_status.xc;
   //dp->current_pose.yc=msg_robot_status.yc;
   dp->current_pose.theta=msg_robot_status.theta;
   dp->steer_ang=msg_robot_status.steering_ang;
   dp->vel=msg_robot_status.linear_speed;

   t_now=(double)(msg_robot_status.header.stamp.sec+msg_robot_status.header.stamp.nsec*1E-9);
   deltaT=t_now-t_ant;
   t_ant=t_now;

   if (dp->running_control==true){//Mission execution in progress
     if (dp->maneouver_pending==true){//Maneouver in progress
       //------------------------ DOOR DETECTION  ---------------------
       if (dp->point2.maneouver==THROUGHT_DOOR_MOVEMENT){//Door crossing
         dp->dist_door_travelled+=msg_robot_status.linear_speed*deltaT;
         //printf("Dist travelled:%f\n",dp->dist_door_travelled);
         dp->detectObstacles(msg_laser);
         dp->detectGaps();

         if (dp->gap_list.size()){
           dp->getGapDirection();
         }else{
           //printf("No gaps\n");
           dp->gap_center=0.0;
         }
       }else{//Other kind of maneouvers (Not door crossing)
         dp->detectObstaculization(msg_laser);
       }
     }else{//Path follow (not maneouver)
       dp->detectObstaculization(msg_laser);
     }
  }
   */
}

int main (int argc, char** argv)
{
  //Construction of the data processor
  dp=new DataProcessor();

  //Read execution arguments
  for (int i=1;i<argc;i++){
    if (!strcmp(argv[i],"-f")){//Load parameters file in YAML format
      dp->ReadConfigurationFile(argv[i+1]);
    }
  }

  //Get the steering range in radians
  dp->steeringSemiRangeRad=dp->steerSemirangePls/dp->plsAng;

  //Connect to ROS
  ros::init(argc, argv, "path_follower_node_eusebio");
  ROS_INFO("Node path_follower_node_eusebio Connected to roscore");

  ros::NodeHandle nh_;//ROS Handler

  //Publications
  //dp->pub_marks = nh_.advertise<sensor_msgs::PointCloud2>("/path_follower/mark", 1);
  //dp->pub_joy_command = nh_.advertise<joy::Joy>("/joy", 1);
  dp->pub_joy_command = nh_.advertise<sensor_msgs::Joy>("/joy", 1);

  dp->pub_path = nh_.advertise<nav_msgs::Path>("/path_follower/path", 1);
  dp->marker_pub = nh_.advertise<visualization_msgs::Marker>("/path_follower/markers", 1);
  dp->pub_status = nh_.advertise<path_follower_node_eusebio::Status>("/path_follower/status", 1);
  //dp->pub_wii_command = nh_.advertise<std_msgs::String>("/wii_irodom/command", 1);

  //Subscribing
  ROS_INFO("Subscribing to topics\n");
  ros::Subscriber command_sub=nh_.subscribe< std_msgs::String> ("/path_follower/command", 1, CommandCallback);
 // ros::Subscriber wii_irodom_sub=nh_.subscribe< wii_irodom::Status> ("/wii_irodom/status", 1, WiiIROdomCallback);

  //Sinchronized topics
#ifdef USE_ROBOT_STATUS
  message_filters::Subscriber<eusebio::RobotStatus> robot_status_sub(nh_,"/eusebio/RobotStatus", 1);
#elif defined USE_KALMAN_FILTER_STATUS
  message_filters::Subscriber<kalman_filter_node::Status> robot_status_sub(nh_,"/kalman_filter/Status", 1);
#endif
  message_filters::Subscriber<sensor_msgs::LaserScan> laser_sub(nh_,"/base_scan" , 1);

#ifdef USE_ROBOT_STATUS
  typedef sync_policies::ApproximateTime<eusebio::RobotStatus, sensor_msgs::LaserScan>  MySyncPolicy1;
#elif defined USE_KALMAN_FILTER_STATUS
  typedef sync_policies::ApproximateTime<kalman_filter_node::Status, sensor_msgs::LaserScan>  MySyncPolicy1;
#endif
  Synchronizer<MySyncPolicy1> sync1(MySyncPolicy1(10), robot_status_sub, laser_sub);
  sync1.registerCallback(&LaserSyncCallback);

  ROS_INFO("Starting ROS event main loop @%f FPS\n",dp->fps);

  //Main loop
  ros::Rate rate(dp->fps);
  while (ros::ok()){
    if((dp->mission_accomplished==false) && (dp->running_control==true)){//There is a mission in progress and its execution is activated
      if (dp->maneouver_pending==true){//Maneouvers
        switch(dp->point2.maneouver){
          case FORWARD_MOVEMENT:
            dp->maneouverExecuted(FORWARD_MOVEMENT);
            break;
          case CHARGER_CONNECT:
            {
              std_msgs::String msg_command;

              dp->stopRobot();//Stop and no more path follow control
              printf("SENDING CONNECT COMMAND\n");
              msg_command.data="CONNECT";
//              dp->pub_wii_command.publish(msg_command);                
            }
            break;
          case THROUGHT_DOOR_MOVEMENT:
            if (fabs(dp->dist_door_travelled)>=MAX_DOOR_TRAVEL_DIST){
              dp->maneouverExecuted(THROUGHT_DOOR_MOVEMENT);
            }
            break;
          case CHARGER_DISCONNECT:
            {
              std_msgs::String msg_command;

              dp->stopRobot();//Stop and no more path follow control
              printf("SENDING DISCONNECT COMMAND\n");
              msg_command.data="DISCONNECT";//Still not disconnected
  //            dp->pub_wii_command.publish(msg_command);      
            }
            break;
        }//switch maneouver
 
      }else{//Following segments
        dp->getTrackError();//Calculates the cross_track_err and the long_track_err and updates the point_reached flag if the 2th point of the segment has been reached

        //printf("CTE:%f OTE:%f LTE:%f\n",dp->cross_track_err,dp->ori_track_err, dp->long_track_err);

        if (dp->long_track_err>=1.0){
           dp->point_reached=true;
           dp->maneouver_pending=true;
           if (dp->path.path[dp->path.idx].maneouver==THROUGHT_DOOR_MOVEMENT){
             dp->dist_door_travelled=0.0;
           }
        }
      }

      dp->calculateReferences(&dp->vel_ref,&dp->steer_ref);

      dp->sendReferences(dp->vel_ref,dp->steer_ref);
    }

    dp->PublishStatus();
    ros::spinOnce();
    rate.sleep();
  }

  //ros::spin();

  delete dp;

  ROS_INFO("ROS-Node Terminated\n");
}
